from io import BytesIO
from uuid import uuid4

import gridfs
from mongoengine import GridFSProxy

from dashserve import JupyterDash
from dashserve.serializer import DashAppSerializer
from omegaml.backends.basedata import BaseDataBackend


class DashAppBackend(BaseDataBackend):
    """ Backend to support storing ``dash`` applications inside notebooks

    .. warning::

        **deprecated** prefer to build a proper Pip package.

    Usage::

        # in Jupyter Notebook
        from dashserve import JupyterDash

        app = JupyterDash(__name__)
        app.layout = html.Div(children=[
           ...
        ]

        # save
        om.scripts.put(app, 'apps/myapp')

    See Also::
        * dashserve - https://github.com/omegaml/dashserve
    """
    KIND = 'python.dash'

    @classmethod
    def supports(cls, obj, name, **kwargs):
        return isinstance(obj, JupyterDash)

    @property
    def _fs(self):
        return self.data_store.fs

    def _serializer(self, app=None):
        return DashAppSerializer(app)

    def put(self, obj, name, attributes=None, **kwargs):
        serialized = self._serializer(obj).serialize(wrap=False)
        bbuf = BytesIO(serialized)
        bbuf.seek(0)
        # see if we have a file already, if so replace the gridfile
        meta = self.data_store.metadata(name)
        if not meta:
            filename = uuid4().hex
            fileid = self._fs.put(bbuf, filename=filename)
            meta = self.data_store._make_metadata(name=name,
                                             prefix=self.data_store.prefix,
                                             bucket=self.data_store.bucket,
                                             kind=DashAppBackend.KIND,
                                             attributes=attributes,
                                             gridfile=GridFSProxy(grid_id=fileid))
        else:
            meta.gridfile.replace(bbuf)
        return meta.save()

    def get(self, name, **kwargs):
        """

        Args:
            name (str): the name of the dash app to load
            **kwargs: the kwargs to pass to DashappServe. Use to override
                      Dash kwargs

        Returns:
            Dashserve app i.e. deserialized dash server

        """
        meta = self.data_store.metadata(name)
        if meta:
            try:
                outf = meta.gridfile
            except gridfs.errors.NoFile as e:
                raise e
            serialized = outf.read()
            deserialized = self._serializer().deserialize(serialized, **kwargs)
        else:
            raise gridfs.errors.NoFile(
                ">{0}< does not exist in scripts bucket '{1}'".format(
                    name, self.data_store.bucket))
        return deserialized






import glob

import mlflow
import os
from tarfile import TarFile

from omegaml.backends.basemodel import BaseModelBackend


class MLFlowModelBackend(BaseModelBackend):
    """ A backend for mlflow models

    This provides storage support for mlflow models. The following
    semantics are supported:

    * mlflow storage format: directories with a MLmodel spec file
    * any of the mlflow-supported frameworks
    * mlflow.models.Model: mlflow custom models of any flavor
    * mlflow.pyfunc.PythonModel:  mlflow Python models of any flavor

    For mlflow storage format, the backend creates a tar file and stores
    it as a blob object.

    For any mlflow-supported frameworks, the backend infers the model
    format and uses the corresponding mlflow.<flavor>.save_model() method

    For mlflow.models.Model instances, the backend uses Model.save() to
    save the model

    For mlflow.pyfunc.PythonModel, the backend uses pyfunc.save_model() to
    save the model

    On loading the model, the backend invariable uses mlflow.pyfunc.load_model()
    to load it back.

    Usage:
        1) Use case: deploy a mlflow model

            # in mlflow save some model
            mlflow.save_model(model, "/path/to/mymodel")
            # deploy using omega-ml
            om.models.put("/path/to/mymodel", 'mymodel')

        2) Use case: deploy a PythonModel

            # mymodel.py
            class MyModel(mlflow.pyfunc.PythonModel):
                def predict(context, data):
                    ...

            model = MyModel()
            om.models.put(model, 'mymodel')

        3) Use case: deploy any mlflow-supported flavor

            model = SomeModel()
            om.models.put(model, 'mymodel', kind='mlflow.model')

        Once the model is saved in om.models, it can be used by
        om.runtime.model() as well as throught the REST API as any
        other model.
    """
    KIND = 'mlflow.model'
    MLFLOW_TYPES = (mlflow.models.Model, mlflow.pyfunc.PythonModel)
    MLFLOW_PREFIX = 'mlflow://'

    @classmethod
    def supports(cls, obj, name, **kwargs):
        if str(obj).startswith(cls.MLFLOW_PREFIX):
            obj = obj.replace(cls.MLFLOW_PREFIX, '')
        is_mlflow_type = isinstance(obj, cls.MLFLOW_TYPES)
        is_mlflow_mlfile = isinstance(obj, str) and obj.lower().endswith('mlmodel') and cls._is_path(cls, obj)
        is_mlflow_modelpath = isinstance(obj, str) and cls._is_path(cls, os.path.join(obj, 'MLmodel'))
        kind_requested = kwargs.get('kind') == cls.KIND
        is_mlflow_flavor, _ = cls._infer_model_flavor(cls, obj)
        return any([is_mlflow_type, is_mlflow_mlfile, is_mlflow_modelpath, kind_requested, is_mlflow_flavor])

    def _infer_model_flavor(self, model):
        from google.protobuf.message import Message
        model_type = type(model).__module__
        model_flavor = model_type.split('.', 1)[0]
        flavor_supported = model_flavor in mlflow._model_flavors_supported
        non_models = (
            lambda m: isinstance(m, Message),
        )
        not_a_model = any(test(model) for test in non_models)
        return not not_a_model and flavor_supported, getattr(mlflow, model_flavor, None)

    def _package_model(self, model, key, tmpfn, **kwargs):
        # package the model using corresponding save or save_model method
        model_path = os.path.join(self.model_store.tmppath, key + '.mlflow')
        if isinstance(model, mlflow.models.Model):
            model.save(model, tmpfn)
        elif isinstance(model, mlflow.pyfunc.PythonModel):
            mlflow.pyfunc.save_model(model_path, python_model=model,
                                     artifacts=kwargs.get('artifacts'))
        elif isinstance(model, str) and (self._is_path(model)
                                         or model.startswith(self.MLFLOW_PREFIX)):
            # a mlflow model local storage
            # https://www.mlflow.org/docs/latest/models.html#storage-format
            if model.startswith(self.MLFLOW_PREFIX):
                model = model.replace(self.MLFLOW_PREFIX, '')
            if model.lower().endswith('mlmodel'):
                model_path = os.path.dirname(model)
            else:
                model_path = model
        else:
            # some supported model flavor perhaps?
            _, flavor = self._infer_model_flavor(model)
            flavor.save_model(model, model_path, **kwargs)
        with TarFile(tmpfn, mode='w') as tarf:
            tarf.add(model_path, recursive=True)
        return tmpfn

    def _extract_model(self, infile, key, tmpfn, **kwargs):
        model_path = os.path.join(self.model_store.tmppath, key + '.mlflow')
        with open(tmpfn, 'wb') as fout:
            fout.write(infile.read())
        with TarFile(tmpfn, mode='r') as tarf:
            tarf.extractall(model_path)
        for fn in glob.glob(f'{model_path}/**/MLmodel', recursive=True):
            model = mlflow.pyfunc.load_model(os.path.dirname(fn))
            break
        return model



from os.path import basename, dirname

import os
import tempfile
from shutil import make_archive, unpack_archive
from subprocess import run

from omegaml.backends.basedata import BaseDataBackend
from omegaml.backends.package.packager import RunnablePackageMixin


class MLFlowProjectBackend(RunnablePackageMixin, BaseDataBackend):
    """
    Backend to support storage of MLFlow projects

    Usage::

        om.scripts.put('mlflow://path/to/MLProject', 'myproject')
        om.scripts.get('myproject')

    See Also:
        https://www.mlflow.org/docs/latest/projects.html#project-directories
    """
    KIND = 'mlflow.project'
    MLFLOW_PREFIX = 'mlflow://'

    @classmethod
    def supports(self, obj, name, **kwargs):
        is_mlflow_prefix = isinstance(obj, str) and obj.startswith(self.MLFLOW_PREFIX)
        is_mlflow_kind = kwargs.get('kind') == self.KIND
        return is_mlflow_kind or is_mlflow_prefix

    def put(self, obj, name, attributes=None, **kwargs):
        """
        save a MLFlow project

        :param obj: full path to the MLFlow project directory
                    mlflow://path/to/MLProject
        :param name: name to store
        :return: the Metadata object
        """
        pkgsrc = obj.split(self.MLFLOW_PREFIX)[-1]
        if os.path.exists(pkgsrc):
            distdir = os.path.join(pkgsrc, 'dist')
            os.makedirs(distdir, exist_ok=True)
            tarfn = os.path.join(distdir, f'{name}')
            pkgdist = make_archive(tarfn, 'gztar', root_dir=pkgsrc, base_dir='.')
        else:
            raise FileNotFoundError(pkgsrc)
        filename = self.data_store.object_store_key(name, 'pkg', hashed=True)
        gridfile = self._store_to_file(self.data_store, pkgdist, filename)
        return self.data_store._make_metadata(
            name=name,
            prefix=self.data_store.prefix,
            bucket=self.data_store.bucket,
            kind=self.KIND,
            attributes=attributes,
            gridfile=gridfile).save()

    def get(self, name, localpath=None, **kwargs):
        """
        Load MLFlow project from store

        This copies the projects's .tar.gz file from om.scripts to a local temp
        path and returns the MLFlowProject to it

        :param name: the name of the package
        :param keep: keep the packages load path in sys.path, defaults to False
        :param localpath: the local path to store the package
        :param install: if True call pip install on the retrieved package
        :param kwargs:
        :return: the loaded module
        """
        pkgname = basename(name)
        dstdir = localpath or self.data_store.tmppath
        packagefname = '{}.tar.gz'.format(os.path.join(localpath or self.packages_path, pkgname))
        os.makedirs(dirname(packagefname), exist_ok=True)
        meta = self.data_store.metadata(name)
        outf = meta.gridfile
        with open(packagefname, 'wb') as pkgf:
            pkgf.write(outf.read())
        unpack_archive(packagefname, dstdir)
        if localpath:
            mod = localpath
        else:
            mod = MLFlowProject(dstdir)
        return mod

    @property
    def packages_path(self):
        return os.path.join(self.data_store.tmppath, 'packages')


class MLFlowProject:
    """ a proxy to the MLFlow project that runs a script

    This provides the mod.run() interface for scripts so that
    we can use the same semantics for mlflow projects and pypi
    packages
    """

    def __init__(self, uri):
        self.uri = uri

    def run(self, om, pure_python=False, conda=True, **kwargs):
        options = ' '.join(f'--{k.replace("_", "-")} {v}' for k, v in kwargs.items())
        if not conda:
            options += ' --no-conda'
        tmpdir = tempfile.mkdtemp()
        # fix issue
        with open(os.path.join(tmpdir, 'activate'), 'w') as fout:
            fout.write('#/bin/bash')
            fout.write('conda activate %1')
        cmd = fr'PATH={tmpdir}:$PATH; cd {tmpdir}; chmod +x ./activate; mlflow run {options} {self.uri}'
        print(cmd)
        output = run(cmd, capture_output=True, shell=True)
        print(output)
        if output.stderr:
            output = {
                'stdout': output.stdout.decode('utf8'),
                'stderr': output.stderr.decode('utf8'),
            }
        else:
            output = output.stdout.decode('utf8')
        return {
            'output': output,
        }



class CombinedOmegaStoreMixin:
    def __init__(self, stores):
        self._stores = stores

    def get(self, name, *args, **kwargs):
        store, store_name = self.store_by_name(name)
        return store.get(store_name, *args, **kwargs)

    def put(self, obj, name, *args, **kwargs):
        store, store_name = self.store_by_name(name)
        return store.put(obj, store_name, *args, **kwargs)

    def list(self, pattern=None, regexp=None, raw=False, **kwargs):
        index = []
        pattern = pattern or ''
        for store in self._stores:
            regexp = regexp.replace(store.prefix) if regexp else None
            # list by name|pattern, return list of Metadata or list of names
            members = store.list(pattern.replace(store.prefix, '', 1),
                                 regexp=regexp, raw=raw, **kwargs)
            # either add obj:Metadata as is, or obj:str as 'prefix/name'
            add = lambda obj: f'{store.prefix}{obj}' if not raw else obj
            index.extend(add(obj) for obj in members)
        return index

    def help(self, name_or_obj):
        import sys
        import pydoc

        if isinstance(name_or_obj, str):
            try:
                store, store_name = self.store_by_name(name_or_obj)
            except ValueError as e:
                pass
            else:
                return store.help(store_name)
        return help(name_or_obj) if sys.flags.interactive else pydoc.render_doc(name_or_obj, renderer=pydoc.plaintext)

    def metadata(self, name, **kwargs):
        store, store_name = self.store_by_name(name)
        return store.metadata(store_name, **kwargs)

    def stores_prefixes(self):
        return [s.prefix for s in self._stores]

    def store_by_prefix(self, prefix):
        prefix = prefix.replace('datasets/', 'data/')
        for store in self._stores:
            if store.prefix == prefix:
                return store
        raise ValueError(f'there is no store for {prefix}')

    def store_by_meta(self, meta):
        return self.store_by_prefix(meta.prefix)

    def store_by_name(self, name):
        for store in self._stores:
            if name.startswith(store.prefix):
                return store, name.replace(store.prefix, '')
        prefixes = [s.prefix for s in self._stores]
        raise ValueError(f'there is no store for {name}, try prefixing with any of {prefixes}')



import flask
from flask_restx import Model, fields
from werkzeug.exceptions import BadRequest

import omegaml as om


class StrictModel(Model):
    # To implement a model that supports strict validation
    # on fields, we need to explicitly add into the schema
    # 'additionalProperties: False'
    # See: https://github.com/noirbizarre/flask-restplus/issues/241
    @property
    def _schema(self):
        old = super(StrictModel, self)._schema
        old['additionalProperties'] = False
        return old


class AnyObject(fields.Wildcard):
    def __init__(self, *args, **kwargs):
        super().__init__(fields.Raw, **kwargs)


class strict(object):
    # a poor man's stand-in for api.model
    def __init__(self, api):
        self.api = api

    def model(self, name=None, model=None, mask=None, **kwargs):
        # create a strict model and attach to api as in flask-restplus issue #241
        smodel = StrictModel(name, model, mask=mask)
        smodel.__apidoc__.update(kwargs)
        self.api.models[smodel.name] = smodel
        return smodel


class OmegaResourceMixin(object):
    """
    helper mixin to resolve the request to a configured Omega instance
    """

    def __init__(self, *args, **kwargs):
        self._omega_instance = None
        super().__init__(*args, **kwargs)

    def dispatch_request(self, *args, **kwargs):
        self._omega_instance = None  # always start with a fresh omega instance
        return super().dispatch_request(*args, **kwargs)

    @property
    def _omega(self):
        if self._omega_instance is None:
            bucket = flask.request.headers.get('bucket')
            self._omega_instance = om.setup()[bucket]
        return self._omega_instance

    def get_query_payload(self):
        from omegaml.restapi.resources import omega_api
        query = flask.request.args.to_dict()
        payload = omega_api.payload or {}
        return query, payload

    def check_object_authorization(self, pattern):
        from omegaml.restapi import resource_filter
        if resource_filter and not any(rx.match(pattern) for rx in resource_filter):
            return False
        return True

    def create_response_from_resource(self, generic_resource, resource_method, resource_name, resource_pk, *args,
                                      **kwargs):
        query, payload = self.get_query_payload()
        async_body = {
            resource_name: resource_pk,
            'result': 'pending',
        }
        pattern = rf'{resource_name}/{resource_pk}/{resource_method}/'
        if not self.check_object_authorization(pattern):
            raise BadRequest(f'{pattern} is not available')
        try:
            meth = self._get_resource_method(generic_resource, resource_method)
            result = meth(resource_pk, query, payload)
            resp = self.create_maybe_async_response(result, async_body=async_body, **kwargs)
        except Exception as e:
            raise BadRequest(repr(e))
        return resp

    def _get_resource_method(self, resource_name, method_name):
        resource = getattr(self, resource_name)
        meth = getattr(resource, method_name)
        return meth

    @property
    def _generic_model_resource(self):
        from omegaml.backends.restapi.model import GenericModelResource
        return GenericModelResource(self._omega, is_async=self.is_async)

    @property
    def _generic_script_resource(self):
        from omegaml.backends.restapi.script import GenericScriptResource
        return GenericScriptResource(self._omega, is_async=self.is_async)

    @property
    def _generic_service_resource(self):
        from omegaml.backends.restapi.service import GenericServiceResource
        return GenericServiceResource(self._omega, is_async=self.is_async)

    @property
    def _generic_job_resource(self):
        from omegaml.backends.restapi.job import GenericJobResource
        return GenericJobResource(self._omega, is_async=self.is_async)

    @property
    def celeryapp(self):
        return self._omega.runtime.celeryapp



import numpy as np

def run(om, *args, state=None, results=None, **kwargs):
    """
    the script API execution entry point
    :return: result
    """
    for result in results:
        if isinstance(result, np.ndarray):
            om.datasets.put(result, 'callback_results', as_pydata=True)
    return state



import os
import subprocess


def run(om, *args, package=None, file=None, action='install', options=None, **kwargs):
    """ install packages in user environment

    Installation:
        om scripts put envinstall envinstall

    Usage:
        # install
        om datasets put requirements.txt .system/requirements.txt
        om runtime script envinstall

        # uninstall
        om runtime script envinstall run action=uninstall
    """
    lockfile = '/app/pylib/user/envinstall.lock'
    reqfile = '.system/requirements.txt'
    package = package or ' '
    try:
        with open(lockfile, 'x') as flock:
            # only run if we can get an exclusive lock
            reqfn = '/tmp/envreq.txt'
            if file:
                with open(reqfn, 'wb') as fout:
                    reqs = om.scripts.get(reqfile).read()
                    fout.write(reqs)
                reqspec = f'-r {reqfn}'
            else:
                reqspec = package
            action_opts = {
                'install': f'-U --user {reqspec}',
                'uninstall': f'-y {reqspec}'
            }
            opts = options or action_opts.get(action) or ''
            pipcmd = f'pip {action} {opts}'
            process = subprocess.run(pipcmd.split(' '),
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.STDOUT,
                                     encoding='utf-8')
    except IOError as e :
        result = f'another envinstall process is running on this node: {e}'
    else:
        result = str(process.stdout)
    finally:
        try:
            os.remove(lockfile)
        except:
            pass
    return result



from time import sleep

import platform

import pandas as pd
import unittest
from sklearn.datasets import load_iris
from sklearn.linear_model import LogisticRegression

from omegaml import Omega
from omegaml.backends.experiment import ExperimentBackend, OmegaSimpleTracker, OmegaProfilingTracker
from omegaml.documents import Metadata
from omegaml.tests.util import OmegaTestMixin


class TrackingTestCases(OmegaTestMixin, unittest.TestCase):
    def setUp(self):
        self.om = om = Omega()
        self.clean()
        om.models.register_backend(ExperimentBackend.KIND, ExperimentBackend)

    def test_ensure_active(self):
        # explicit start
        om = self.om
        exp = om.runtime.experiment('test')
        with self.assertRaises(ValueError):
            exp.log_param('foo', 'bar')
        run = exp.start()
        exp.log_param('foo', 'bar')
        data = exp.data(run=run)
        self.assertIsNotNone(run)
        self.assertEqual(len(data), 2)
        # reuse latest run
        exp = om.runtime.experiment('test')
        with self.assertRaises(ValueError):
            exp.log_param('foo', 'bar')
        exp.use()
        exp.log_param('foo', 'bax')
        exp.stop()
        data = exp.data(run=run)
        self.assertIsNotNone(run)
        self.assertEqual(len(data), 4)
        # implied run should be NOT set in with block if use() was called previously
        # rationale: with exp should always start a new run
        with exp as xexp:
            xexp.log_param('foo', 'bar')
        data = xexp.data()
        self.assertEqual(len(data), 3)
        self.assertEqual(set(data['event']), {'start', 'stop', 'param'})
        self.assertEqual(data['run'].iloc[-1], run + 1)

    def test_simple_tracking(self):
        # create a model
        om = self.om
        iris = load_iris()
        X = iris.data
        Y = iris.target
        lr = LogisticRegression(solver='liblinear', multi_class='auto')
        lr.fit(X, Y)
        # run a local experiment and track its result
        with om.runtime.experiment('myexp') as exp:
            # check exp is associated with correct stores
            self.assertEqual(exp._model_store, om.models)
            self.assertEqual(exp._store, om.datasets)
            score = lr.score(X, Y)
            exp.log_metric('accuracy', score)
        # check the experiemnt is logged
        self.assertIn('experiments/myexp', om.models.list('experiments/*'))
        exp = om.models.get('experiments/myexp', data_store=om.datasets)
        data = exp.data(event='metric')
        self.assertIsInstance(exp, OmegaSimpleTracker)
        self.assertIsNotNone(data)
        self.assertEqual(len(data), 1)
        self.assertEqual(data.iloc[0]['key'], 'accuracy')
        self.assertEqual(data.iloc[0]['value'], score)
        # get back the tracker as an object
        tracker = om.models.get('experiments/myexp', raw=True, data_store=om.datasets)
        self.assertIsInstance(tracker, OmegaSimpleTracker)
        # check remote execution of experiement
        # -- we want to easily transfer code execution to a remote
        # -- while keeping the code semantics as local as possible
        om.models.put(lr, 'mymodel')
        with om.runtime.experiment('myexp', provider='simple') as exp:
            om.runtime.model('mymodel').fit(X, Y)
            om.runtime.model('mymodel').score(X, Y)
            exp.log_metric('accuracy', 1)
            exp.log_artifact(lr, 'mymodel')
            exp.log_artifact(om.models.metadata('mymodel'), 'mymodel_meta')
        tracker = om.models.get('experiments/myexp', raw=True, data_store=om.datasets)
        data = tracker.data()
        self.assertIsInstance(data, pd.DataFrame)
        self.assertEqual(len(data), 15)  # includes runtime task events
        self.assertEqual(len(data[data.event == 'start']), 2)
        self.assertEqual(len(data[data.event == 'stop']), 2)
        self.assertEqual(len(data[data.event == 'artifact']), 4)
        artifacts = data[data.event == 'artifact']['value'].to_list()
        obj = exp.restore_artifact(key='mymodel')
        self.assertIsInstance(obj, LogisticRegression)
        obj = exp.restore_artifact(key='mymodel_meta')
        self.assertIsInstance(obj, Metadata)
        # related is stored by runtime automatically as the delegate's metadata
        obj = exp.restore_artifact(key='related')
        self.assertIsInstance(obj, Metadata)
        # check that the current metadata lists the experiment for tracability
        meta = om.models.metadata('mymodel')
        self.assertIn('myexp', meta.attributes['tracking']['experiments'])

    def test_tracking_from_metadata(self):
        # create a model
        om = self.om
        iris = load_iris()
        X = iris.data
        Y = iris.target
        lr = LogisticRegression(solver='liblinear', multi_class='auto')
        lr.fit(X, Y)
        om.models.put(lr, 'mymodel')
        # no tracking
        tracker = om.runtime.experiment('expfoo')
        exp = tracker.experiment
        data = exp.data()
        self.assertIsNone(data)
        # implicit tracking via metadata
        om.models.put(lr, 'mymodel', attributes={
            'tracking': {
                'default': 'expfoo2',
            }})
        om.runtime.model('mymodel').score(X, Y)
        tracker = om.runtime.experiment('expfoo2')
        exp = tracker.experiment
        data = exp.data()
        self.assertIsNotNone(data)
        self.assertEqual(len(data), 6)

    def test_tracking_predictions_explicit_metadata(self):
        # create a model
        om = self.om
        iris = load_iris()
        X = iris.data
        Y = iris.target
        lr = LogisticRegression(solver='liblinear', multi_class='auto')
        lr.fit(X, Y)
        om.models.put(lr, 'mymodel', attributes={
            'tracking': {
                'default': 'expfoo'
            }
        })
        om.runtime.model('mymodel').score(X, Y)
        tracker = om.runtime.experiment('expfoo')
        exp = tracker.experiment
        data = exp.data()
        self.assertIsNotNone(data)
        self.assertEqual(len(data), 6)

    def test_tracking_predictions_implicit_tracker(self):
        # create a model
        om = self.om
        iris = load_iris()
        X = iris.data
        Y = iris.target
        lr = LogisticRegression(solver='liblinear', multi_class='auto')
        lr.fit(X, Y)
        om.models.put(lr, 'mymodel')
        tracker = om.runtime.experiment('expfoo')
        tracker.track('mymodel')
        om.runtime.model('mymodel').score(X, Y)
        exp = tracker.experiment
        data = exp.data()
        self.assertIsNotNone(data)
        self.assertEqual(len(data), 6)
        # run a prediction to see if this is tracked
        om.runtime.model('mymodel').predict(X)
        tracker = om.runtime.experiment('expfoo')
        data = exp.data()
        self.assertIsNotNone(data)
        self.assertEqual(len(data), 11)

    def test_tracking_runtime_taskid(self):
        # create a model
        om = self.om
        iris = load_iris()
        X = iris.data
        Y = iris.target
        lr = LogisticRegression(solver='liblinear', multi_class='auto')
        lr.fit(X, Y)
        om.models.put(lr, 'mymodel')
        tracker = om.runtime.experiment('expfoo')
        tracker.track('mymodel')
        resp = om.runtime.model('mymodel').score(X, Y)
        exp = tracker.experiment
        data = exp.data(taskid=resp.task_id)
        self.assertIsNotNone(data)
        self.assertEqual(len(data), 6)
        self.assertEqual(data['run'].unique(), [1])
        resp = om.runtime.model('mymodel').score(X, Y)
        data = exp.data(taskid=resp.task_id)
        self.assertEqual(len(data), 6)
        self.assertEqual(data['run'].unique(), [2])
        # run a prediction to see if this is tracked
        om.runtime.model('mymodel').predict(X)
        tracker = om.runtime.experiment('expfoo')
        data = exp.data()
        self.assertIsNotNone(data)
        self.assertEqual(len(data), 3 * 6 -1 ) # 3 runs, last one has no metric

    def test_empty_experiment_data(self):
        om = self.om
        with om.runtime.experiment('myexp') as exp:
            pass
        exp = om.models.get('experiments/myexp', data_store=om.datasets)
        # we have at least a 'system' event
        self.assertEqual(len(exp.data(event='metric')), 0)
        self.assertEqual(len(exp.data()), 2)  # start, stop events

    def test_experiment_explicit_logging(self):
        om = self.om
        with om.runtime.experiment('myexp') as exp:
            exp.log_metric('accuracy', .98)
        exp = om.models.get('experiments/myexp', data_store=om.datasets)
        self.assertEqual(len(exp.data(event='metric')), 1)

    def test_experiments_not_versioned(self):
        om = self.om
        with om.runtime.experiment('myexp') as exp:
            pass
        meta = om.models.metadata('experiments/myexp')
        self.assertNotIn('versions', meta.attributes)

    def test_experiment_artifact(self):
        om = self.om
        with om.runtime.experiment('myexp') as exp:
            exp.log_artifact(dict(test='data'), 'foo')
        # by name
        data = exp.restore_artifact('foo')
        self.assertIsInstance(data, dict)
        # by value
        data = exp.data(key='foo')
        data = exp.restore_artifact(value=data.iloc[0].value)

    def test_notrack(self):
        # create a model
        om = self.om
        iris = load_iris()
        X = iris.data
        Y = iris.target
        lr = LogisticRegression(solver='liblinear', multi_class='auto')
        lr.fit(X, Y)
        # use runtime to fit model without tracking
        om.models.put(lr, 'mymodel')
        om.runtime.model('mymodel').fit(X, Y)
        tracker = om.runtime.experiment('exp')
        exp = tracker.experiment
        # no tracking, no data
        data = exp.data()
        self.assertIsNone(data)

    def test_experiment_existing_model(self):
        om = self.om
        iris = load_iris()
        X = iris.data
        Y = iris.target
        lr = LogisticRegression(solver='liblinear', multi_class='auto')
        om.models.put(lr, 'foo')
        with om.runtime.experiment('experiments/foo') as exp:
            exp.log_metric(5, 'accuracy')
        self.assertIn('experiments/foo', om.models.list())

    def test_drop_experiment(self):
        om = self.om
        # create experiment, add some data
        with om.runtime.experiment('foo') as exp:
            exp.log_metric(5, 'accuracy')
        self.assertEqual(len(exp.data()), 3)
        # check experiment and data where created
        self.assertIn('experiments/foo', om.models.list())
        meta = om.models.metadata('experiments/foo')
        dataset = meta.attributes['tracking']['dataset']
        self.assertIsNotNone(dataset)
        self.assertIn(dataset, om.datasets.list(hidden=True))
        self.assertIn('experiments/foo', om.models.list())
        # add a metric, then clean
        exp.log_metric(5, 'accuracy')
        self.assertEqual(len(exp.data()), 4)
        # explicit drop
        om.models.drop('experiments/foo', data_store=om.datasets)
        self.assertNotIn(dataset, om.datasets.list(hidden=True))
        # no more data present
        self.assertIsNone(exp.data())

    def test_log_system(self):
        om = self.om
        with om.runtime.experiment('test') as exp:
            exp.log_system()
        data = exp.data(key='system')
        system = data.iloc[0]['value']
        self.assertEqual(system['platform'], platform.uname()._asdict())
        self.assertEqual(system['platform']['node'], platform.node())

    def test_profiling(self):
        om = self.om
        with om.runtime.experiment('proftest', provider='profiling') as exp:
            exp.profiler.interval = 0.1
            sleep(1.5)
        data = exp.data(event='profile')
        self.assertGreaterEqual(len(data), 9)
        xexp = om.runtime.experiment('proftest')
        self.assertIsInstance(xexp.experiment, OmegaProfilingTracker)

    def test_track_then_notracking(self):
        # test experiments are stopped afterwards
        om = self.om
        # start an experiment
        with om.runtime.experiment('myexp') as exp:
            pass
        exp = om.models.get('experiments/myexp', data_store=om.datasets)
        # check that it is removed afterwards
        om.runtime.ping()
        self.assertEqual(len(exp.data(event='metric')), 0)
        self.assertEqual(len(exp.data()), 2)  # includes start, stop


if __name__ == '__main__':
    unittest.main()



from unittest import TestCase

from nbformat import v4
from urllib.parse import quote

from omegaml import Omega, restapi
from omegaml.client.auth import OmegaRestApiAuth
from omegaml.tests.core.restapi.util import RequestsLikeTestClient
from omegaml.tests.util import OmegaTestMixin


class JobResourceTests(OmegaTestMixin, TestCase):
    def setUp(self):
        app = restapi.create_app()
        self.client = RequestsLikeTestClient(app, is_json=True)
        self.om = Omega()
        self.auth = OmegaRestApiAuth('user', 'pass')
        self.clean()

    def tearDown(self):
        pass

    def url(self, pk=None, action=None, query=None):
        url = '/api/v1/job/'
        if pk is not None:
            url += '{pk}/'.format(pk=quote(pk, safe='')) # encode / in name
        if action is not None:
            url += '{action}'.format(**locals())
        if query is not None:
            url += '?{query}'.format(**locals())
        return url

    def test_get_job_info(self):
        om = self.om
        # create a notebook
        cells = []
        code = "print('hello')"
        cells.append(v4.new_code_cell(source=code))
        notebook = v4.new_notebook(cells=cells)
        # put notebook
        meta = om.jobs.put(notebook, 'testjob')
        resp = self.client.get(self.url('testjob'))
        self.assertHttpOK(resp)
        # see if we get job data
        data = self.deserialize(resp)
        self.assertIn('created', data)
        self.assertIn('job_results', data)
        self.assertIn('job_runs', data)
        self.assertIn('job', data)
        self.assertEqual(data['job'], 'testjob.ipynb')
        # run the job locally and see if we get results ok
        om.jobs.run('testjob')
        resp = self.client.get(self.url('testjob'))
        self.assertHttpOK(resp)
        data = self.deserialize(resp)
        self.assertIn('created', data)
        self.assertEqual(len(data['job_results']), 1)
        self.assertEqual(len(data['job_runs']), 1)
        # get job info for results
        resultsjob = data['job_results'][0]
        resp = self.client.get(self.url(resultsjob))
        self.assertHttpOK(resp)
        data = self.deserialize(resp)
        self.assertIn('source_job', data)

    def test_job_run(self):
        om = self.om
        # create a notebook
        cells = []
        code = "print('hello')"
        cells.append(v4.new_code_cell(source=code))
        notebook = v4.new_notebook(cells=cells)
        # put notebook
        meta = om.jobs.put(notebook, 'testjob')
        # run the job on the cluster
        resp = self.client.post(self.url('testjob', action='run'), json={})
        self.assertHttpOK(resp)
        data = self.deserialize(resp)
        self.assertIn('created', data)
        self.assertEqual(len(data['job_results']), 1)
        self.assertEqual(len(data['job_runs']), 1)
        # get job info for results
        resultsjob = data['job_results'][0]
        resp = self.client.get(self.url(resultsjob))
        self.assertHttpOK(resp)
        data = self.deserialize(resp)
        self.assertIn('source_job', data)

    def test_job_run_async(self):
        om = self.om
        # create a notebook
        cells = []
        code = "print('hello')"
        cells.append(v4.new_code_cell(source=code))
        notebook = v4.new_notebook(cells=cells)
        # put notebook
        meta = om.jobs.put(notebook, 'testjob')
        # run the job on the cluster
        resp = self.client.post(self.url('testjob', action='run'),
                                headers=self._async_headers, json={})
        resp = self._check_async(resp)
        self.assertHttpOK(resp)
        data = self.deserialize(resp)['response']
        self.assertIn('created', data)
        self.assertEqual(len(data['job_results']), 1)
        self.assertEqual(len(data['job_runs']), 1)
        # get job info for results
        resultsjob = data['job_results'][0]
        resp = self.client.get(self.url(resultsjob))
        self.assertHttpOK(resp)
        data = self.deserialize(resp)
        self.assertIn('source_job', data)





try:
    import tensorflow as tf
except ModuleNotFoundError:
    pass
else:
    from .tfestimatormodel import TFEstimatorModelBackend, TFEstimatorModel
    from .tfkeras import TensorflowKerasBackend
    from .tfkerassavedmodel import TensorflowKerasSavedModelBackend
    from .tfsavedmodel import TensorflowSavedModelPredictor, TensorflowSavedModelBackend, ServingInput

    # compatibility with previous tensorflow versions
    FN_MAP = {}
    if tf.__version__.startswith('1.'):
        FN_MAP['pandas_input_fn'] = getattr(tf.estimator.inputs, 'pandas_input_fn')
        FN_MAP['numpy_input_fn'] = getattr(tf.estimator.inputs, 'numpy_input_fn')
        FN_MAP['convert_to_tensor']  = getattr(tf.compat.v1, 'convert_to_tensor')
    elif tf.__version__.startswith('2.'):
        FN_MAP['pandas_input_fn'] = getattr(tf.compat.v1.estimator.inputs, 'pandas_input_fn')
        FN_MAP['numpy_input_fn'] = getattr(tf.compat.v1.estimator.inputs, 'numpy_input_fn')
        FN_MAP['convert_to_tensor'] = getattr(tf, 'convert_to_tensor')


    def _tffn(name):
        return FN_MAP[name]



def run(om, payload, *args, **query):
    """
    the script API execution entry point
    :return: result
    """
    import pandas as pd
    factor = payload.get('factor')
    df = pd.DataFrame({
        'a': range(0, 5),
        'b': range(0, 5),
    })
    return (df * factor).to_dict(orient='list')



from os.path import basename

import os
import re

from omegaml.backends.basedata import BaseDataBackend
from omegaml.backends.package.packager import install_and_import, RunnablePackageMixin


class PythonPipSourcedPackageData(RunnablePackageMixin, BaseDataBackend):
    """
    Backend to support locally sourced custom scripts deployment to runtimes cluster

    This supports any pip-installable remote source like git, hg, svn, bzr

    Usage::

        # pypi package
        om.scripts.put('pypi://<pkgname>', '<pkgname>')

        # git hosted
        om.scripts.put('git+https://github.com/account/repo/<pkgname>', '<pkgname>')

        # retrieve the package and call pip install
        om.scripts.get('name')

    Notes:
        * `pypi://<pkgname>` can be any specification supported by pip, e.g.::

            om.scripts.put('pypi://<pkgname==version>', ...)
            om.scripts.put('git+https://....@tag#egg=<pkgname>', ...)

        * the package specification is stored as ``metadata.kind_meta['pip_source']``

    See Also:
        * https://packaging.python.org/tutorials/packaging-projects/
    """
    KIND = 'pipsrc.package'

    @classmethod
    def supports(self, obj, name, **kwargs):
        # see https://pip.pypa.io/en/stable/reference/pip_install/#vcs-support
        # pypi is our own protocol to designate pypi hosted packages
        # all others are pip supported formats
        # the pattern supports <source>:// and <source>+<protocol>:// formats
        # for pypi, .get() will remove the pypi:// part
        pip_protocols = '|'.join(('git', 'hg', 'svn', 'bzr', 'pypi'))
        pattern = r'^{}(\+.+)?://.*'
        is_pip_protocol = lambda v: re.match(pattern.format(pip_protocols), v)
        return isinstance(obj, str) and is_pip_protocol(obj)

    def put(self, obj, name, attributes=None, **kwargs):
        """
        save a python package

        This stores a PIP-sourceable package specification in the objects
        ``Metadata.kind_meta``::

             { 'pip_source': '<obj:package spec>' }

        :param obj: full path to package file or directory, syntax as
                    pkg://path/to/dist.tar.gz or pkg://path/to/setup.py
        :param name: name of package. must be the actual name of the package
                     as specified in setup.py
        :return: the Metadata object
        """
        kind_meta = {
            'pip_source': obj,
        }
        return self.data_store._make_metadata(
            name=name,
            prefix=self.data_store.prefix,
            bucket=self.data_store.bucket,
            kind=PythonPipSourcedPackageData.KIND,
            kind_meta=kind_meta,
            attributes=attributes).save()

    def get(self, name, keep=False, **kwargs):
        """
        Load package from store, install it locally and load.

        This runs `pip install <package spec>`, taking the package specification
        from Metadata.kind_meta['pip-source']. See the put() method for details.

        :param name: the name of the package
        :param keep: keep the packages load path in sys.path, defaults to False
        :param kwargs:
        :return: the loaded module
        """
        pkgname = basename(name)
        meta = self.data_store.metadata(name)
        dstdir = self.packages_path
        pip_source = meta.kind_meta['pip_source']
        # pip does not know about pypi
        pip_name = pip_source.replace('pypi://', '')
        mod = install_and_import(pip_name, pkgname, dstdir, keep=keep)
        return mod


    @property
    def packages_path(self):
        return os.path.join(self.data_store.tmppath, 'packages')



import joblib

from omegaml.backends.basecommon import BackendBaseCommon
from omegaml.util import reshaped


class BaseModelBackend(BackendBaseCommon):
    """
    OmegaML BaseModelBackend to be subclassed by other arbitrary backends

    This provides the abstract interface for any model backend to be implemented
    Subclass to implement custom backends.

    Essentially a model backend:

     * provides methods to serialize and deserialize a machine learning model for a given ML framework
     * offers fit() and predict() methods to be called by the runtime
     * offers additional methods such as score(), partial_fit(), transform()

    Model backends are the middleware that connects the om.models API to specific frameworks. This class
    makes it simple to implement a model backend by offering a common syntax as well as a default implementation
    for get() and put().

    Methods to implement:
        # for model serialization (mandatory)
        @classmethod supports() - determine if backend supports given model instance
        _package_model() - serialize a model instance into a temporary file
        _extract_model() - deserialize the model from a file-like

        By default BaseModelBackend uses joblib.dumps/loads to store the model as serialized
        Python objects. If this is not sufficient or applicable to your type models, override these
        methods.

        Both methods provide readily set up temporary file names so that all you have to do is actually
        save the model to the given output file and restore the model from the given input file, respectively.
        All other logic has already been implemented (see get_model and put_model methods).

        # for fitting and predicting (mandatory)
        fit()
        predict()

        # other methods (optional)
        fit_transform() - fit and return a transformed dataset
        partial_fit() - fit incrementally
        predict_proba() - predict probabilities
        score() - score fitted classifier vv test dataset

    """
    _backend_version_tag = '_om_backend_version'
    _backend_version = '1'

    def __init__(self, model_store=None, data_store=None, tracking=None, **kwargs):
        assert model_store, "Need a model store"
        assert data_store, "Need a data store"
        self.model_store = model_store
        self.data_store = data_store
        self.tracking = tracking

    @classmethod
    def supports(self, obj, name, **kwargs):
        """
        test if this backend supports this obj
        """
        return False

    @property
    def _call_handler(self):
        # the model store handles _pre and _post methods in self.perform()
        return self.model_store

    def get(self, name, **kwargs):
        """
        retrieve a model

        :param name: the name of the object
        :param version: the version of the object (not supported)
        """
        # support new backend architecture while keeping back compatibility
        return self.get_model(name, **kwargs)

    def put(self, obj, name, **kwargs):
        """
        store a model

        :param obj: the model object to be stored
        :param name: the name of the object
        :param attributes: attributes for meta data
        """
        # support new backend architecture while keeping back compatibility
        return self.put_model(obj, name, **kwargs)

    def drop(self, name, force=False, version=-1, **kwargs):
        return self.model_store._drop(name, force=force, version=version)

    def _package_model(self, model, key, tmpfn, **kwargs):
        """
        implement this method to serialize a model to the given tmpfn

        Args:
            model:
            key:
            tmpfn:
            **kwargs:

        Returns:
            tmpfn or absolute path of serialized file
        """
        with open(tmpfn, 'wb') as outf:
            joblib.dump(model, outf)
        return tmpfn


    def _extract_model(self, infile, key, tmpfn, **kwargs):
        """
        implement this method to deserialize a model from the given infile

        Args:
            infile: this is a file-like object supporting read() and seek(). if
                deserializing from this does not work directly, use tmpfn
            key:
            tmpfn:
            **kwargs:

        Returns:
            model instance
        """
        obj = joblib.load(infile)
        return obj

    def get_model(self, name, version=-1, **kwargs):
        """
        Retrieves a pre-stored model
        """
        meta = self.model_store.metadata(name)
        storekey = self.model_store.object_store_key(name, 'omm', hashed=True)
        model = self._extract_model(meta.gridfile, storekey,
                                    self._tmp_packagefn(self.model_store, storekey), **kwargs)
        return model

    def put_model(self, obj, name, attributes=None, _kind_version=None, **kwargs):
        """
        Packages a model using joblib and stores in GridFS
        """
        storekey = self.model_store.object_store_key(name, 'omm', hashed=True)
        tmpfn = self._tmp_packagefn(self.model_store, storekey)
        packagefname = self._package_model(obj, storekey, tmpfn, **kwargs) or tmpfn
        gridfile = self._store_to_file(self.model_store, packagefname, storekey)
        kind_meta = {
            self._backend_version_tag: self._backend_version,
        }
        return self.model_store._make_metadata(
            name=name,
            prefix=self.model_store.prefix,
            bucket=self.model_store.bucket,
            kind=self.KIND,
            kind_meta=kind_meta,
            attributes=attributes,
            gridfile=gridfile).save()

    def predict(
          self, modelname, Xname, rName=None, pure_python=True, **kwargs):
        """
        predict using data stored in Xname

        :param modelname: the name of the model object
        :param Xname: the name of the X data set
        :param rName: the name of the result data object or None
        :param pure_python: if True return a python object. If False return
            a dataframe. Defaults to True to support any client.
        :param kwargs: kwargs passed to the model's predict method
        :return: return the predicted outcome
        """
        model = self.model_store.get(modelname)
        data = self.data_store.get(Xname)
        if not hasattr(model, 'predict'):
            raise NotImplementedError
        result = model.predict(reshaped(data))
        if pure_python:
            result = result.tolist()
        if rName:
            meta = self.data_store.put(result, rName)
            result = meta
        return result


    def predict_proba(
          self, modelname, Xname, rName=None, pure_python=True, **kwargs):
        """
        predict the probability using data stored in Xname

        :param modelname: the name of the model object
        :param Xname: the name of the X data set
        :param rName: the name of the result data object or None
        :param pure_python: if True return a python object. If False return
            a dataframe. Defaults to True to support any client.
        :param kwargs: kwargs passed to the model's predict method
        :return: return the predicted outcome
        """
        raise NotImplementedError

    def fit(self, modelname, Xname, Yname=None, pure_python=True, **kwargs):
        """
        fit the model with data

        :param modelname: the name of the model object
        :param Xname: the name of the X data set
        :param Yname: the name of the Y data set
        :param pure_python: if True return a python object. If False return
           a dataframe. Defaults to True to support any client.
        :param kwargs: kwargs passed to the model's predict method
        :return: return the meta data object of the model
        """
        raise NotImplementedError

    def partial_fit(
          self, modelname, Xname, Yname=None, pure_python=True, **kwargs):
        """
        partially fit the model with data (online)

        :param modelname: the name of the model object
        :param Xname: the name of the X data set
        :param Yname: the name of the Y data set
        :param pure_python: if True return a python object. If False return
           a dataframe. Defaults to True to support any client.
        :param kwargs: kwargs passed to the model's predict method
        :return: return the meta data object of the model
        """

        raise NotImplementedError

    def fit_transform(
          self, modelname, Xname, Yname=None, rName=None, pure_python=True,
          **kwargs):
        """
        fit and transform using data

        :param modelname: the name of the model object
        :param Xname: the name of the X data set
        :param Yname: the name of the Y data set
        :param rName: the name of the transforms's result data object or None
        :param pure_python: if True return a python object. If False return
           a dataframe. Defaults to True to support any client.
        :param kwargs: kwargs passed to the model's transform method
        :return: return the meta data object of the model
        """
        raise NotImplementedError

    def transform(self, modelname, Xname, rName=None, **kwargs):
        """
        transform using data

        :param modelname: the name of the model object
        :param Xname: the name of the X data set
        :param rName: the name of the transforms's result data object or None
        :param kwargs: kwargs passed to the model's transform method
        :return: return the transform data of the model
        """
        raise NotImplementedError

    def score(
          self, modelname, Xname, Yname=None, rName=True, pure_python=True,
          **kwargs):
        """
        score using data

        :param modelname: the name of the model object
        :param Xname: the name of the X data set
        :param Yname: the name of the Y data set
        :param rName: the name of the transforms's result data object or None
        :param pure_python: if True return a python object. If False return
           a dataframe. Defaults to True to support any client.
        :param kwargs: kwargs passed to the model's predict method
        :return: return the score result
        """
        raise NotImplementedError





import json
import unittest
from pathlib import Path
from unittest import TestCase

import warnings

from omegaml import Omega
from omegaml.tests.util import OmegaTestMixin
from omegaml.util import module_available

try:
    from omegaml.backends.mlflow.gitprojects import MLFlowGitProjectBackend
    from omegaml.backends.mlflow.localprojects import MLFlowProjectBackend, MLFlowProject
except:
    warnings.warn("mlflow not installed")
else:
    @unittest.skipUnless(module_available('mlflow'), 'mlflow projects not available CURRENTLY DISABLED')
    class TestMLFlowProjects(OmegaTestMixin, TestCase):
        def setUp(self):
            om = self.om = Omega()
            self.clean()
            om.models.register_backend(MLFlowProjectBackend.KIND, MLFlowProjectBackend)
            om.models.register_backend(MLFlowGitProjectBackend.KIND, MLFlowGitProjectBackend)

        def test_mlflow_project_local(self):
            import omegaml
            # store a local MLFlow project, use kind=mlflow.project
            project_path = Path(omegaml.__file__).parent / 'example' / 'mlflow' / 'project'
            om = self.om
            meta = om.scripts.put(str(project_path), 'myproject', kind='mlflow.project')
            self.assertEqual(meta.kind, MLFlowProjectBackend.KIND)
            # get it back
            mod = om.scripts.get('myproject')
            self.assertIsInstance(mod, MLFlowProject)
            # run it on omegaml runtime
            result = om.runtime.script('myproject').run(entry_point='main.py', conda=False).get()
            data = json.loads(result)
            self.assertIn('output', data['result'])
            self.assertIn('hello', data['result']['output']['stdout'])

        def test_mlflow_project_local_prefix(self):
            import omegaml
            # store a local MLFlow project, use mlflow:// prefix
            project_path = Path(omegaml.__file__).parent / 'example' / 'mlflow' / 'project'
            om = self.om
            meta = om.scripts.put('mlflow://' + str(project_path), 'myproject')
            self.assertEqual(meta.kind, MLFlowProjectBackend.KIND)
            # get it back
            mod = om.scripts.get('myproject')
            self.assertIsInstance(mod, MLFlowProject)
            # run it on omegaml runtime
            result = om.runtime.script('myproject').run(entry_point='main.py', conda=False).get()
            data = json.loads(result)
            self.assertIn('output', data['result'])
            self.assertIn('hello', data['result']['output']['stdout'])

        @unittest.skip('enable once issue #258 is fixed')
        def test_mlflow_project_remote_https(self):
            # store a local MLFlow project
            project_path = 'https://github.com/omegaml/omegaml#omegaml/example/mlflow/project'
            om = self.om
            meta = om.scripts.put(str(project_path), 'myproject', kind='mlflow.gitproject')
            self.assertEqual(meta.kind, MLFlowGitProjectBackend.KIND)
            # get it back
            mod = om.scripts.get('myproject')
            self.assertIsInstance(mod, MLFlowProject)
            # run it on omegaml runtime
            result = om.runtime.script('myproject').run(entry_point='mlflow_tracking.py', conda=False).get()
            data = json.loads(result)
            self.assertIn('output', data['result'])
            self.assertIn('succeeded', data['result']['output']['stderr'])

        @unittest.skip('enable once issue #258 is fixed')
        def test_mlflow_gitproject_remote_ssh(self):
            # store a local MLFlow project
            project_path = 'mlflow+ssh://git@github.com/mlflow/mlflow.git#examples/quickstart'
            om = self.om
            meta = om.scripts.put(str(project_path), 'myproject')
            self.assertEqual(meta.kind, MLFlowGitProjectBackend.KIND)
            # get it back
            mod = om.scripts.get('myproject')
            self.assertIsInstance(mod, MLFlowProject)
            # run it on omegaml runtime
            result = om.runtime.script('myproject').run(entry_point='mlflow_tracking.py', conda=False).get()
            data = json.loads(result)
            self.assertIn('output', data['result'])
            self.assertIn('succeeded', data['result']['output']['stderr'])

        @unittest.skip('enable once issue #258 is fixed')
        def test_mlflow_gitproject_remote_https(self):
            # store a local MLFlow project
            project_path = 'mlflow+https://github.com/mlflow/mlflow.git#examples/quickstart'
            om = self.om
            meta = om.scripts.put(str(project_path), 'myproject')
            self.assertEqual(meta.kind, MLFlowGitProjectBackend.KIND)
            # get it back
            mod = om.scripts.get('myproject')
            self.assertIsInstance(mod, MLFlowProject)
            # run it on omegaml runtime
            result = om.runtime.script('myproject').run(entry_point='mlflow_tracking.py', conda=False).get()
            data = json.loads(result)
            self.assertIn('output', data['result'])
            self.assertIn('succeeded', data['result']['output']['stderr'])




#!/usr/bin/env python
from setuptools import setup

setup(name='callback', version='1.0', description='results saving callback demo', author='omegaml',
      author_email='info@omegaml.io', url='http://omegaml.io', packages=['callback'], )



import os
import yaml

from omegaml.client.auth import AuthenticationEnv
from omegaml.client.userconf import ensure_api_url
from omegaml.client.util import protected
from omegaml.omega import Omega as CoreOmega
from omegaml.runtimes import OmegaRuntime


class OmegaCloud(CoreOmega):
    """
    Client API to omegaml cloud

    Provides the following APIs:

    * :code:`datasets` - access to datasets stored in the cluster
    * :code:`models` - access to models stored in the cluster
    * :code:`runtimes` - access to the cluster compute resources
    * :code:`jobs` - access to jobs stored and executed in the cluster
    * :code:`scripts` - access to lambda modules stored and executed in the cluster

    """

    def __init__(self, auth=None, **kwargs):
        """
        Initialize the client API
        """
        self.auth = auth
        super(OmegaCloud, self).__init__(**kwargs)

    def _make_runtime(self, celeryconf):
        return OmegaCloudRuntime(self, bucket=self.bucket, defaults=self.defaults, celeryconf=celeryconf)

    def _clone(self, **kwargs):
        kwargs.update(auth=self.auth)
        return super()._clone(**kwargs)

    def __repr__(self):
        return 'OmegaCloud(bucket={}, auth={})'.format(self.bucket or 'default', repr(self.auth))


class OmegaCloudRuntime(OmegaRuntime):
    """
    omegaml hosted compute cluster gateway
    """

    def __init__(self, omega, **kwargs):
        super().__init__(omega, **kwargs)
        self._auth_kwarg = protected('auth')

    def __repr__(self):
        return 'OmegaCloudRuntime(auth={})'.format(repr(self.omega.auth))

    @property
    def _common_kwargs(self):
        common = super()._common_kwargs
        common['task'].update({self._auth_kwarg: self.auth.token})
        label = common['routing'].get('label', 'default')
        if getattr(self.omega.defaults, 'OMEGA_TASK_ROUTING_ENABLED', False):
            common['routing'].update(self.build_account_routing(label))
        return common

    @property
    def auth(self):
        return self.omega.auth

    @property
    def dispatch_label(self):
        return self.omega.auth.userid

    def build_account_routing(self, label):
        # build the queue
        account_routing = {
            'label': '{self.dispatch_label}-{label}'.format(**locals())
        }
        return account_routing


def setup(userid=None, apikey=None, api_url=None, qualifier=None, bucket=None,
          view=None, make_default=True):
    import omegaml as om_mod
    api_url = ensure_api_url(api_url, om_mod._base_config)
    view = (view if view is not None
            else getattr(om_mod._base_config, 'OMEGA_SERVICES_INCLUSTER', False))
    auth_env = AuthenticationEnv.secure()
    om = auth_env.get_omega_from_apikey(userid=userid, apikey=apikey,
                                        api_url=api_url, qualifier=qualifier, view=view)
    # prepare om to be linked to default
    om.Omega = OmegaCloud
    om.setup = lambda *args, **kwargs: setup(
        **{**dict(userid=userid, apikey=apikey, qualifier=qualifier, bucket=bucket),
           **kwargs})
    om._om = om
    # ensure link to deferred instance
    om_mod.Omega = OmegaCloud
    om_mod.link(om) if make_default else None
    return om[bucket]


def setup_from_config(config_file=None, fallback=None):
    """ setup from a cloud configuration file

    If the configuration files does not contain OMEGA_USERID and OMEGA_APIKEY,
    will use given fallback.
    """
    from omegaml import _base_config
    config_file = config_file or _base_config.OMEGA_CONFIG_FILE
    AuthenticationEnv.secure()
    if isinstance(config_file, str) and os.path.exists(config_file):
        with open(config_file, 'r') as fin:
            userconfig = yaml.safe_load(fin)
            if isinstance(userconfig, dict) and 'OMEGA_USERID' in userconfig:
                try:
                    omega = setup(userid=userconfig['OMEGA_USERID'],
                                  apikey=userconfig['OMEGA_APIKEY'],
                                  qualifier=userconfig.get('OMEGA_QUALIFIER'),
                                  api_url=userconfig.get('OMEGA_RESTAPI_URL'))
                except Exception as e:
                    # TODO make this a SystemError so that OmegaDeferredIstance.setup reverts to proper defaults
                    raise SystemError(f'Could not login using config file {config_file}, error={e}')
            elif fallback:
                # if alternative loading was provided, use that
                omega = fallback()
            else:
                raise SystemError('No cloud cloud userid/apikey found in config file {}.'.format(config_file))
            return omega
    raise SystemError('Config file {} does not exist'.format(config_file))



class GenericScriptResource(object):
    def __init__(self, om, is_async=False):
        self.om = om
        self.is_async = is_async

    def is_eager(self):
        return getattr(self.om.runtime.celeryapp.conf, 'CELERY_ALWAYS_EAGER', False)

    def run(self, script_id, query, payload):
        """
        Args:
            script_id (str): the name of the model
            query (dict): the query parameters
            payload (dict): the json body

        Returns:
            result
        """
        om = self.om
        payload = {} if payload is None else payload
        promise = om.runtime.script(script_id).run(payload, __format='python', **query)
        result = promise if self.is_async else self.prepare_result(promise.get(), resource_name=script_id)
        return result

    def prepare_result(self, result, resource_name=None, **kwargs):
        return result



import re

from omegaml.mdataframe import MDataFrame


class LazyGetMixin:
    """
    OmegaStore mixin to support chunked lazy get via name

    Usage:

        equivalent of ``om.datasets.get('foo', lazy=True).iterchunks()``::

            mdf = om.datasets.get('foo#')
            mdf = om.datasets.get('foo#iterchunks')

        equivalent of ``om.datasets.get('foo', lazy=True).iterchunks(chunksize=10)``::

            mdf = om.datasets.get('foo#iterchunks:chunksize=10')

        equivalent of ``om.datasets.get('foo', lazy=True).iloc[0:10]``::

            mdf = om.datasets.get('foo#rows:start=1,end=10')
    """
    # requires a trailing ; to work in all cases, see https://regex101.com/r/lYeKAw/1
    ops_pattern = re.compile(r"(?P<name>.*)#(?P<opspec>.*?);(.*)$")

    def metadata(self, name, *args, **kwargs):
        if isinstance(name, str):
            name, opspec = self._extract_opspec(name)
        return super().metadata(name, *args, **kwargs)

    def get(self, name, *args, **kwargs):
        name, opspec = self._extract_opspec(name or '')
        if opspec is not None:
            kwargs = {**kwargs, **dict(lazy=True)}
            lazy = super().get(name, *args, **kwargs)
            if ':' in opspec:
                op, op_kwargs_specs = opspec.split(':', 1)
                op_kwargs = {}
                for kw in op_kwargs_specs.split(','):
                    k, v = kw.split('=')
                    op_kwargs[k] = v
            else:
                op, op_kwargs = self._default_op(name, lazy)
            meth = getattr(lazy, op, lambda *args, **kwargs: value)
            value = meth(**op_kwargs)
        else:
            value = super().get(name, *args, **kwargs)
        return value

    def _extract_opspec(self, name):
        match = self.ops_pattern.match(name + ';') if '#' in name else None
        opspec = None
        if match is not None:
            name, opspec, _ = match.groups()
        return name, opspec

    def _default_op(self, name, lazy):
        if isinstance(lazy, MDataFrame):
            op = 'iterchunks'
            opkwargs = {}
        else:
            op = '__repr__'
            opkwargs = {}
        return op, opkwargs



from multiprocessing import Process, Queue
from unittest import TestCase

from time import sleep

from omegaml import Omega
from omegaml.util import delete_database

try:
    from minibatch import Buffer, connectdb, stream
    from minibatch.contrib.omegaml import DatasetSink
except ModuleNotFoundError:
    print("minibatch not available - not testing streams")
else:
    class MiniBatchTests(TestCase):
        def setUp(self):
            delete_database()
            self.om = Omega()
            db = self.om.datasets.mongodb
            self.url = self.om.mongo_url + '?authSource=admin'
            connectdb(url=self.url)

        def test_stream(self):
            """
            Test a stream writes to a buffer
            """
            from minibatch import stream
            om = self.om
            om.datasets.mongodb
            s = stream('test', url=self.url)
            s.append({'foo': 'bar1'})
            s.append({'foo': 'bar2'})
            count = len(list(doc for doc in Buffer.objects.all()))
            self.assertEqual(count, 2)

        def test_fixed_size(self):
            """
            Test batch windows of fixed sizes work ok
            """
            from minibatch import streaming, stream

            def consumer(q, url):
                # note the stream decorator blocks the consumer and runs the decorated
                # function asynchronously upon the window criteria is satisfied
                om = Omega(mongo_url=url)

                @streaming('test', size=2, url=url, keep=True, queue=q,
                           sink=DatasetSink(om, 'consumer'))
                def myprocess(window):
                    return {'myprocess': True, 'data': window.data}

            # start stream and consumer
            s = stream('test', url=self.url)
            q = Queue()
            proc = Process(target=consumer, args=(q, self.url))
            proc.start()
            # fill stream
            for i in range(10):
                s.append({'index': i})
            # give it some time to process
            sleep(5)
            q.put(True)
            proc.join()
            # expect 5 entries, each of length 2
            windows = list(doc for doc in self.om.datasets.collection('consumer').find())
            self.assertEqual(len(windows), 5)
            self.assertTrue(all(len(w['data']) == 2 for w in windows))

        def test_timed_window_relaxed(self):
            """
            Test timed windows work ok
            """
            from minibatch import streaming

            def consumer(q, url):
                # note the stream decorator blocks the consumer and runs the decorated
                # function asynchronously upon the window criteria is satisfied
                om = Omega(mongo_url=url)

                @streaming('test', interval=1, keep=True, url=url, queue=q,
                           relaxed=True,
                           sink=DatasetSink(om, 'consumer'))
                def myprocess(window):
                    return {'myprocess': True, 'data': window.data}

            # start stream and consumer
            q = Queue()
            s = stream('test', url=self.url)
            proc = Process(target=consumer, args=(q, self.url,))
            proc.start()
            # fill stream
            for i in range(10):
                s.append({'index': i})
                sleep(1)
            # give it some time to process
            sleep(2)
            q.put(True)
            proc.join()
            # expect at least 5 entries (10 x .5 = 5 seconds), each of length 1-2
            windows = list(doc for doc in self.om.datasets.collection('consumer').find())
            self.assertGreater(len(windows), 5)
            # most windows have one or more entries
            self.assertTrue(sum(len(w['data']) >= 1 for w in windows) >= 4)
            # all messages were processed
            self.assertEqual(sum(len(w['data']) for w in windows), 10)

        def test_timed_window_fixed(self):
            """
            Test relaxed time windows work ok
            """
            from minibatch import streaming

            def consumer(q, url):
                # note the stream decorator blocks the consumer and runs the decorated
                # function asynchronously upon the window criteria is satisfied
                om = Omega(mongo_url=url)

                @streaming('test', interval=1, relaxed=False,
                           keep=True, url=url, queue=q,
                           sink=DatasetSink(om, 'consumer'))
                def myprocess(window):
                    return {'myprocess': True, 'data': window.data}

            # start stream and consumer
            s = stream('test', url=self.url)
            q = Queue()
            proc = Process(target=consumer, args=(q, self.url,))
            proc.start()
            # fill stream
            for i in range(10):
                s.append({'index': i})
                sleep(1)
            # give it some time to process
            sleep(1)
            q.put(True)
            proc.join()
            # expect at least 5 entries (10 x .5 = 5 seconds), each of length 1-2
            windows = list(doc for doc in self.om.datasets.collection('consumer').find())
            self.assertGreater(len(windows), 5)
            # most windows have one or more entries
            self.assertTrue(sum(len(w['data']) >= 1 for w in windows) >= 4)
            # all messages were processed
            self.assertEqual(sum(len(w['data']) for w in windows), 10)



import os

from omegaml.restapi.app import create_app

if __name__ == '__main__':
    port = os.environ.get('OMEGA_WEB_PORT')
    app = create_app()
    app.run(host='0.0.0.0', port=port, debug=True)



import getpass

import dill
import os
import pandas as pd
import pkg_resources
import platform
from base64 import b64encode, b64decode
from datetime import datetime
from itertools import product
from uuid import uuid4

from omegaml.backends.basemodel import BaseModelBackend
from omegaml.documents import Metadata
from omegaml.util import _raise, settings


class ExperimentBackend(BaseModelBackend):
    """ ExperimentBackend provides storage of tracker configurations

    Usage:

        To log metrics and other data::

            with om.runtime.experiment('myexp') as exp:
                om.runtime.model('mymodel').fit(X, Y)
                om.runtime.model('mymodel').score(X, Y) # automatically log score result
                exp.log_metric('mymetric', value)
                exp.log_param('myparam', value)
                exp.log_artifact(X, 'X')
                exp.log_artifact(Y, 'Y')
                exp.log_artifact(om.models.metadata('mymodel'), 'mymodel')

        To log data and automatically profile system data::

            with om.runtime.experiment('myexp', provider='profiling') as exp:
                om.runtime.model('mymodel').fit(X, Y)
                om.runtime.model('mymodel').score(X, Y) # automatically log score result
                exp.log_metric('mymetric', value)
                exp.log_param('myparam', value)
                exp.log_artifact(X, 'X')
                exp.log_artifact(Y, 'Y')
                exp.log_artifact(om.models.metadata('mymodel'), 'mymodel')

            # profiling data contains metrics for cpu, memory and disk use
            data = exp.data(event='profile')

        To get back experiment data without running an experiment::

            # recommended way
            exp = om.runtime.experiment('myexp').use()
            exp_df = exp.data()

            # experiments exist in the models store
            exp = om.models.get('experiments/myexp')
            exp_df = exp.data()

    See Also:

        * :class:`omegaml.backends.experiment.OmegaSimpleTracker`
        * :class:`omegaml.backends.experiment.OmegaProfilingTracker`
    """
    KIND = 'experiment.tracker'
    exp_prefix = 'experiments/'

    @classmethod
    def supports(self, obj, name, **kwargs):
        return isinstance(obj, TrackingProvider)

    def put(self, obj, name, **kwargs):
        name = f'{self.exp_prefix}{name}' if not name.startswith('experiments') else name
        # FIXME use proper pickle magic to avoid storing the store
        store = obj._store
        obj._store = None
        obj._model_store = None
        meta = super().put(obj, name, **kwargs)
        meta.attributes.setdefault('tracking', {})
        meta.attributes['tracking']['dataset'] = obj._data_name
        meta.save()
        obj._store = store
        obj._model_store = self.model_store
        return meta

    def get(self, name, raw=False, data_store=None, **kwargs):
        assert data_store is not None, "experiments require a datastore, specify data_store=om.datasets"
        tracker = super().get(name, **kwargs)
        name = os.path.basename(name)
        tracker._store = data_store
        tracker._model_store = self.model_store
        return tracker.experiment(name) if not raw else tracker

    def drop(self, name, force=False, version=-1, data_store=None, **kwargs):
        data_store = data_store or self.data_store
        meta = self.model_store.metadata(name)
        dataset = meta.attributes.get('tracking', {}).get('dataset')
        data_store.drop(dataset, force=True) if dataset else None
        return self.model_store._drop(name, force=force, version=version, **kwargs)


class TrackingProvider:
    """ TrackingProvider implements an abstract interface to experiment tracking

    Concrete implementations like MLFlow, Sacred or Neptune.ai can be implemented
    based on TrackingProvider. In combination with the runtime's OmegaTrackingProxy
    this provides a powerful tracking interface that scales with your needs.

    How it works:

        1. Experiments created using ``om.runtime.experiment()`` are stored as
           instances of a TrackingProvider concrete implementation

        2. Upon retrieval of an experiment, any call to its API is proxied to the
           actual implementation, e.g. MLFlow

        3. On calling a model method via the runtime, e.g. ``om.runtime.model().fit()``,
           the TrackingProvider information is passed on to the runtime worker,
           and made available as the backend.tracking property. Thus within a
           model backend, you can always log to the tracker by using::

                with self.tracking as exp:
                    exp.log_metric() # call any TrackingProvider method

        4. omega-ml provides the OmegaSimpleTracker, which implements a tracking
           interface similar to packages like MLFlow, Sacred. See ExperimentBackend
           for an example.
    """

    def __init__(self, experiment, store=None, model_store=None):
        self._experiment = experiment
        self._run = None
        self._store = store
        self._model_store = model_store
        self._extra_log = None

    @property
    def userid(self):
        defaults = getattr(self._store, 'defaults', None) or settings()
        return getattr(defaults, 'OMEGA_USERID', getpass.getuser())

    def experiment(self, name=None):
        self._experiment = name or self._experiment
        return self

    def track(self, obj, store=None, label=None):
        """ attach this experiment to the named object

        Usage:

            # use in experiment context
            with om.runtime.experiment('myexp') as exp:
                exp.track('mymodel')

            # use on experiment directly
            exp = om.runtime.experiment('myexp')
            exp.track('mymodel')

        Args:
                obj (str): the name of the object
                store (OmegaStore): optional, om.models, om.scripts, om.jobs.
                    If not provided will use om.models
                label (str): optional, the label of the worker, default is
                    'default'

        Note:
                This modifies the object's metadata.attributes::

                    { 'tracking': { label: self._experiment }
        """
        label = label or 'default'
        store = store or self._model_store
        meta = store.metadata(obj)
        meta.attributes.setdefault('tracking', {})
        meta.attributes['tracking'] = {
            label: self._experiment
        }
        meta.save()
        return meta

    def active_run(self):
        self._run = (self._run or 0) + 1
        return self._run

    def status(self, run=None):
        return 'STOPPED'

    def start(self):
        raise NotImplementedError

    def stop(self):
        raise NotImplementedError

    def log_event(self, event, key, value, step=None, **extra):
        raise NotImplementedError

    def log_metric(self, key, value, step=None, **extra):
        raise NotImplementedError

    def log_artifact(self, obj, name, step=None, **extra):
        raise NotImplementedError

    def log_param(self, key, value, step=None, **extra):
        raise NotImplementedError

    def log_extra(self, remove=False, **kwargs):
        # add extra logging information for every subsequent log_xyz call
        raise NotImplementedError

    def tensorflow_callback(self):
        return TensorflowCallback(self)

    def data(self, experiment=None, run=None, event=None, step=None, key=None, raw=False):
        raise NotImplementedError

    @property
    def _data_name(self):
        return f'.experiments/{self._experiment}'


class NoTrackTracker(TrackingProvider):
    """ A default tracker that does not record anything """

    def start(self):
        pass

    def stop(self):
        pass

    def log_artifact(self, obj, name, step=None, **extra):
        pass

    def log_metric(self, key, value, step=None, **extra):
        pass

    def log_param(self, key, value, step=None, **extra):
        pass

    def log_event(self, event, key, value, step=None, **extra):
        pass

    def log_extra(self, **kwargs):
        pass

    def data(self, experiment=None, run=None, event=None, step=None, key=None, raw=False):
        pass


class OmegaSimpleTracker(TrackingProvider):
    """ A tracking provider that logs to an omegaml dataset

    Usage::

        with om.runtime.experiment(provider='default') as exp:
            ...
            exp.log_metric('accuracy', .78)
    """
    _provider = 'simple'
    _experiment = None
    _startdt = None
    _stopdt = None

    _ensure_active = lambda self, r: r if r is not None else _raise(
        ValueError('no active run, call .start() or .use() '))

    def active_run(self, run=None):
        """ set the lastest run as the active run

        Args:
            run (int|str): optional or unique task id, if None the
               latest active run will be set, or

        Returns:
            current run (int)
        """
        if run is None:
            latest = self._latest_run
            latest_is_active = (latest is not None and self.status(run=latest) == 'STARTED')
            self._run = latest if latest_is_active else self.start(run=None)
        else:
            self._run = run
        self._experiment = self._experiment or uuid4().hex
        return self._run

    def use(self, run=None):
        """ reuse the latest run instead of starting a new one

        semantic sugar for self.active_run()

        Returns:
            self
        """
        self.active_run(run=run)
        return self

    @property
    def _latest_run(self):
        data = self.data(event='start', raw=True)
        run = data[-1]['run'] if data is not None and len(data) > 0 else None
        return run

    def status(self, run=None):
        """ status of a run

        Args:
            run (int): the run number, defaults to the currently active run

        Returns:
            status in 'STARTED', 'STOPPED'
        """
        self._run = run or self._run or self._latest_run
        data = self.data(event=('start', 'stop'), run=self._run, raw=True)
        no_runs = data is None or len(data) == 0
        has_stop = sum(1 for row in (data or []) if row.get('event') == 'stop')
        return 'PENDING' if no_runs else 'STOPPED' if has_stop else 'STARTED'

    def start(self, run=None):
        """ start a new run

        This starts a new run and logs the start event
        """
        self._run = run or (self._latest_run or 0) + 1
        self._startdt = datetime.utcnow()
        data = self._common_log_data('start', key=None, value=None, step=None, dt=self._startdt)
        self._write_log(data)
        return self._run

    def stop(self):
        """ stop the current run

        This stops the current run and records the stop event
        """
        self._stopdt = datetime.utcnow()
        data = self._common_log_data('stop', key=None, value=None, step=None, dt=self._stopdt)
        self._write_log(data)

    def _common_log_data(self, event, key, value, step=None, dt=None, **extra):
        data = {
            'experiment': self._experiment,
            'run': self._ensure_active(self._run),
            'step': step,
            'event': event,
            'key': key or event,
            'value': value,
            'dt': dt or datetime.utcnow(),
            'node': os.environ.get('HOSTNAME', platform.node()),
            'userid': self.userid,
        }
        data.update(extra)
        data.update(self._extra_log) if self._extra_log else None
        return data

    def _write_log(self, data):
        self._store.put(data, self._data_name, noversion=True)

    def log_artifact(self, obj, name, step=None, **extra):
        """ log any object to the current run

        Usage::

            # log an artifact
            exp.log_artifact(mydict, 'somedata')

            # retrieve back
            mydict_ = exp.restore_artifact('somedata')

        Args:
            obj (obj): any object to log
            name (str): the name of artifact
            step (int): the step, if any
            **extra: any extra data to log

        Notes:
            * bool, str, int, float, list, dict are stored as ``format=type``
            * Metadata is stored as ``format=metadata``
            * objects supported by ``om.models`` are stored as ``format=model``
            * objects supported by ``om.datasets`` are stored as ``format=dataset``
            * all other objects are pickled and stored as ``format=pickle``
        """
        if isinstance(obj, (bool, str, int, float, list, dict)):
            format = 'type'
            rawdata = obj
        elif isinstance(obj, Metadata):
            format = 'metadata'
            rawdata = obj.to_json()
        elif self._model_store.get_backend_byobj(obj) is not None:
            objname = uuid4().hex
            meta = self._model_store.put(obj, f'.experiments/.artefacts/{objname}')
            format = 'model'
            rawdata = meta.name
        elif self._store.get_backend_by_obj(obj) is not None:
            objname = uuid4().hex
            meta = self._store.put(obj, f'.experiments/.artefacts/{objname}')
            format = 'dataset'
            rawdata = meta.name
        else:
            try:
                rawdata = b64encode(dill.dumps(obj)).decode('utf8')
                format = 'pickle'
            except TypeError as e:
                rawdata = repr(obj)
                format = 'repr'
        value = {
            'name': name,
            'data': rawdata,
            'format': format
        }
        data = self._common_log_data('artifact', name, value, step=step, **extra)
        self._write_log(data)

    def log_event(self, event, key, value, step=None, dt=None, **extra):
        data = self._common_log_data(event, key, value, step=step, dt=dt, **extra)
        self._write_log(data)

    def log_param(self, key, value, step=None, dt=None, **extra):
        """ log an experiment parameter

        Args:
            key (str): the parameter name
            value (str|float|int|bool|dict): the parameter value
            step (int): the step
            **extra: any other values to store with event

        Notes:
            * logged as ``event=param``
        """
        data = self._common_log_data('param', key, value, step=step, dt=dt, **extra)
        self._write_log(data)

    def log_metric(self, key, value, step=None, dt=None, **extra):
        """ log a metric value

        Args:
            key (str): the metric name
            value (str|float|int|bool|dict): the metric value
            step (int): the step
            **extra: any other values to store with event

        Notes:
            * logged as ``event=metric``
        """
        data = self._common_log_data('metric', key, value, step=step, dt=dt, **extra)
        self._write_log(data)

    def log_system(self, key=None, value=None, step=None, dt=None, **extra):
        """ log system data

        Args:
            key (str): the key to use, defaults to 'system'
            value (str|float|int|bool|dict): the parameter value
            step (int): the step
            **extra: any other values to store with event

        Notes:
            * logged as ``event=system``
            * logs platform, python version and list of installed packages
        """
        key = key or 'system'
        value = value or {
            'platform': platform.uname()._asdict(),
            'python': '-'.join((platform.python_implementation(),
                                platform.python_version())),
            'packages': ['=='.join((d.project_name, d.version))
                         for d in pkg_resources.working_set]
        }
        data = self._common_log_data('system', key, value, step=step, dt=dt, **extra)
        self._write_log(data)

    def log_extra(self, remove=False, **kwargs):
        """ add additional log information for every subsequent logging call """
        self._extra_log = {} if self._extra_log is None else self._extra_log
        if not remove:
            self._extra_log.update(kwargs)
        else:
            from collections import deque as consume
            deletions = (self._extra_log.pop(k, None) for k in kwargs)
            consume(deletions, maxlen=0)

    def data(self, experiment=None, run=None, event=None, step=None, key=None, raw=False,
             **extra):
        """ build a dataframe of all stored data

        Args:
            experiment (str): the name of the experiment, defaults to its current value
            run (int|list): the run(s) to get data back, defaults to current run, use 'all' for all
            event (str|list): the event(s) to include
            step (int|list): the step(s) to include
            key (str|list): the key(s) to include
            raw (bool): if True returns the raw data instead of a DataFrame

        Returns:
            * data (DataFrame) if raw == False
            * data (list of dicts) if raw == True
            * None if no data exists
        """
        filter = {}
        experiment = experiment or self._experiment
        run = run or self._run
        valid = lambda s: s is not None and str(s).lower() != 'all'
        op = lambda s: {'$in': list(s)} if isinstance(s, (list, tuple)) else s
        if valid(experiment):
            filter['data.experiment'] = op(experiment)
        if valid(run):
            filter['data.run'] = op(run)
        if valid(event):
            filter['data.event'] = op(event)
        if valid(step):
            filter['data.step'] = op(step)
        if valid(key):
            filter['data.key'] = op(key)
        for k, v in extra.items():
            if valid(k):
                filter[f'data.{k}'] = op(v)
        data = self._store.get(self._data_name, filter=filter)
        if data is not None and not raw:
            data = pd.DataFrame.from_records(data)
            data.sort_values('dt', inplace=True) if 'dt' in data.columns else None
        return data

    def restore_artifact(self, key=None, experiment=None, run=None, step=None, value=None):
        """ restore a logged artificat

        Args:
            key (str): the name of the artifact as provided in log_artifact
            run (int): the run for which to query, defaults to current run
            step (int): the step for which to query, defaults to all steps in run
            value (dict): this value is used instead of querying data, use to
               retrieve an artifact from contents of ``.data()``

        Notes:
            * this will restore the artifact according to its type assigned
              by ``.log_artifact()``. If the type cannot be determined, the
              actual data is returned
        """
        if value is None:
            data = self.data(experiment=experiment, run=run, event='artifact', step=step, key=key, raw=True)
            data = data[-1]['value']
        else:
            data = value
        if data['format'] == 'type':
            obj = data['data']
        elif data['format'] == 'metadata':
            meta = self._store._Metadata
            obj = meta.from_json(data['data'])
        elif data['format'] == 'dataset':
            obj = self._store.get(data['data'])
        elif data['format'] == 'model':
            obj = self._model_store.get(data['data'])
        elif data['format'] == 'pickle':
            obj = dill.loads(b64decode((data['data']).encode('utf8')))
        else:
            obj = data.get('data', data)
        return obj


class OmegaProfilingTracker(OmegaSimpleTracker):
    """ A metric tracker that runs a system profiler while the experiment is active

    Will record ``profile`` events that contain cpu, memory and disk profilings.
    See BackgroundProfiler.profile() for details of the profiling metrics collected.

    Usage:

        To log metrics and system performance data::

            with om.runtime.experiment('myexp', provider='profiling') as exp:
                ...

            data = exp.data(event='profile')

    Properties::

        exp.profiler.interval = n.m # interval of n.m seconds to profile, defaults to 3 seconds
        exp.profiler.metrics = ['cpu', 'memory', 'disk'] # all or subset of metrics to collect
        exp.max_buffer = n # number of items in buffer before tracking

    Notes:
        - the profiling data is buffered to reduce the number of database writes, by
          default the data is written on every 6 profiling events (default: 6 * 10 = every 60 seconds)
        - the step reported in the tracker counts the profiling event since the start, it is not
          related to the step (epoch) reported by e.g. tensorflow
        - For every step there is a ``event=profile``, ``key=profile_dt`` entry which you can use
          to relate profiling events to a specific wall-clock time.
        - It usually sufficient to report system metrics in intervals > 10 seconds since machine
          learning algorithms tend to use CPU and memory over longer periods of time.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.profile_logs = []
        self.max_buffer = 6

    def log_profile(self, data):
        """ the callback for BackgroundProfiler """
        self.profile_logs.append(data)
        if len(self.profile_logs) >= (self.max_buffer or 1):
            self.flush()

    def flush(self):
        for step, data in enumerate(self.profile_logs):
            # record the actual time instead of logging time (avoid buffering delays)
            dt = data.get('profile_dt')
            for k, v in data.items():
                self.log_event('profile', k, v, step=step, dt=dt)
        self.profile_logs = []

    def start(self):
        self.profiler = BackgroundProfiler(callback=self.log_profile)
        self.profiler.start()
        super().start()

    def stop(self):
        self.profiler.stop()
        self.flush()
        super().stop()


try:
    from tensorflow import keras
except:
    pass
else:
    class TensorflowCallback(keras.callbacks.Callback):
        """ A callback for Tensorflow Keras models

        Implements the callback protocol according to Tensorflow Keras
        semantics and linking to a :class:`omegaml.backends.experiment.TrackingProvider`

        See Also:

            * https://www.tensorflow.org/guide/keras/custom_callback
        """

        #
        def __new__(cls, *args, **kwargs):
            # generate methods as per specs
            for action, phase in product(['train', 'test', 'predict'], ['begin', 'end']):
                cls.wrap(f'on_{action}_{phase}', 'on_global')
                cls.wrap(f'on_{action}_batch_{phase}', 'on_batch')
            for phase in ['begin', 'end']:
                cls.wrap(f'on_epoch_{phase}', 'on_epoch')
            return super().__new__(cls)

        def __init__(self, tracker):
            self.tracker = tracker
            self.model = None

        def set_model(self, model):
            self.tracker.log_artifact(model, 'model')

        def set_params(self, params):
            for k, v in params.items():
                self.tracker.log_param(k, v)

        def on_global(self, action, logs=None):
            for k, v in (logs or {}).items():
                self.tracker.log_metric(k, v, step=0)

        def on_batch(self, action, batch, logs=None):
            for k, v in (logs or {}).items():
                self.tracker.log_metric(k, v, step=batch)

        def on_epoch(self, action, epoch, logs=None):
            for k, v in (logs or {}).items():
                self.tracker.log_metric(k, v, step=epoch)

        @classmethod
        def wrap(cls, method, fn):
            fn = getattr(cls, fn)

            def inner(self, *args, **kwargs):
                return fn(self, method, *args, **kwargs)

            setattr(cls, method, inner)
            return inner


class BackgroundProfiler:
    """ Profile CPU, Memory and Disk use in a background thread
    """

    def __init__(self, interval=10, callback=print):
        self._stop = False
        self._interval = interval
        self._callback = callback
        self._metrics = ['cpu', 'memory', 'disk']

    def profile(self):
        """
        returns memory and cpu data as a dict
            memory_load (float): percent of total memory used
            memory_total (float): total memory in bytes
            cpu_load (list): cpu load in percent, per cpu
            cpu_count (int): logical cpus,
            cpu_freq (float): MHz of current cpu frequency
            cpu_avg (list): list of cpu load over last 1, 5, 15 minutes
            disk_use (float): percent of disk used
            disk_total (int): total size of disk, bytes
        """
        import psutil
        from datetime import datetime as dt
        p = psutil
        disk = p.disk_usage('/')
        cpu_count = p.cpu_count()
        data = {'profile_dt': dt.utcnow()}
        if 'memory' in self._metrics:
            data.update(memory_load=p.virtual_memory().percent,
                        memory_total=p.virtual_memory().total)
        if 'cpu' in self._metrics:
            data.update(cpu_load=p.cpu_percent(percpu=True),
                        cpu_count=cpu_count,
                        cpu_freq=[f.current for f in p.cpu_freq(percpu=True)],
                        cpu_avg=[x / cpu_count for x in p.getloadavg()])
        if 'disk' in self._metrics:
            data.update(disk_use=disk.percent,
                        disk_total=disk.total)
        return data

    @property
    def interval(self):
        return self._interval

    @interval.setter
    def interval(self, interval):
        self._interval = interval
        self.stop()
        self.start()

    @property
    def metrics(self):
        return self._metrics

    @metrics.setter
    def metrics(self, metrics):
        self._metrics = metrics

    def start(self):
        """ runs a background thread that reports stats every interval seconds

        Every interval, calls callback(data), where data is the output of BackgroundProfiler.profile()
        Stop by BackgroundProfiler.stop()
        """
        import atexit
        from threading import Thread
        from time import sleep

        def runner():
            cb = self._callback
            try:
                while not self._stop:
                    cb(self.profile())
                    sleep(self._interval)
            except (KeyboardInterrupt, SystemExit):
                pass

        # handle exits by stopping the profiler
        atexit.register(self.stop)
        # start the profiler
        self._stop = False
        t = Thread(target=runner)
        t.start()
        # clean up
        atexit.unregister(self.stop)

    def stop(self):
        self._stop = True



import warnings

from functools import reduce

from marshmallow import Schema, fields


class Operation:
    """ apispec operation builder

    Usage:

        spec = APISpec(...)
        operations = [
            Operation(spec, "put")
               .add_request(InputSchema)
               .add_respone(200, OutputSchema),
            Operation(spec, "put")
               .add_request(InputSchema)
               .add_respone(201, OutputSchema),s
        ]
        spec.path('/resource', operations=Operation.to_path(operations))
    """

    def __init__(self, spec, method, description=None, operation_id=None):
        self.spec = spec
        self.method = method
        self.responses = {}
        self.request = None
        self.parameters = {}
        self.description = description
        self.operation_id = operation_id

    def add_response(self, status, schema, description=None):
        """ add operation by http status code

        Args:
            status (int|str): http status code, e.g. 200, 400, "2XX", "4XX"
            schema (Schema|type): the response Schema
            description (str): arbitrary description

        Returns:
            self
        """
        resp = {
            "description": description,
        }
        resp.update(schema=schema) if schema else None
        self.responses[str(status)] = resp
        return self

    def add_request(self, schema):
        """ add a request (body) schema

        Args:
            schema (Schema|type): the name of the schema

        Returns:
            self
        """
        self.request = schema
        return self

    def add_parameter(self, location, name, schema=None, description=None, required=False):
        """ add a parameter in path, header, cookie

        Args:
            location (str): path, header, cookie
            name (str): the name
            schema (Schema): the schema name
            description (str): arbitrary description
            required (bool): whether this parameter is required

        Returns:
            self
        """
        self.parameters[name] = {
            "in": location,
            "schema": schema,
            "description": description,
            "required": required,
        }
        return self

    def to_dict(self):
        """ return the json representation according to the spec major version """
        # swagger 2.0
        if self.request:
            self.add_parameter("body", "body", schema=self.request)
        return {
            self.method: {
                "summary": "summary",
                "description": self.description or ('no description'),
                "operationId": f'{self.operation_id}#{self.method}',
                "consumes": ["application/json"],
                "produces": ["application/json"],
                "parameters": [{
                    "in": param["in"],
                    "name": name,
                    "description": param["description"] or ('no description'),
                    "schema": {
                        "$ref": f'#/definitions/{param["schema"].__name__}'
                    }
                } for name, param in self.parameters.items()],
                "responses": {
                    status: {
                        "description": resp["description"] or ('no description'),
                        "schema": {
                            "$ref": f'#/definitions/{resp["schema"].__name__}'
                        }
                    } for status, resp in self.responses.items()
                }
            }
        }

    @classmethod
    def to_path(self, operations):
        operations = (op.to_dict() for op in operations)
        return reduce(lambda a, b: a.update(b) or a, operations)


class SpecFromResourceHelperBase:
    path_template = '/api/v1/{resource}/{name}/{action}'
    default_actions = {'': ['post']}
    default_orient = 'columns'

    def __init__(self, meta, name, resource, signature, spec, store):
        self.meta = meta
        self.name = name
        self.resource = resource
        self.signature = signature
        self.spec = spec
        self.store = store

    def add_schema_from_datatype(self, name, datatype):
        # idempotent way to add schema.
        # Args:
        #     schema (Schema): the Schema class
        if name in self.spec.components.schemas:
            return
        self.spec.components.schema(name, {}, schema=datatype)

    def add_schema_from_meta(self, name, meta):
        # idempotent way to add schema.
        # Args:
        #     meta (dict): the Metadata.to_dict() equivalent of a dataset
        if name in self.spec.components.schemas and meta:
            return
        self.spec.components.schema(name, {}, meta=meta)

    def match_data_orient(self, datatype, orient):
        # return the datatype as a Nested(Schema) or List(Nested(Schema))
        # -- orient == 'records' => fields.List(fields.Nested(datatype))
        #           == 'columns' => fields.Nested(datatype)
        # Examples:
        #    records => { data: [ { x: 1, y: 2, ... }, ... ]}
        #    columns => { data: { x: [ 1, 2, ... ] }}
        # See also:
        # - https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html
        # - https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_json.html
        return fields.List(fields.Nested(datatype)) if orient == 'records' else fields.Nested(datatype)

    def render(self):
        # render endpoint
        orient = self.signature.get('orient') or self.default_orient
        datatypes = self.datatypes(orient)
        InputSchema = datatypes['X']
        DefaultOutputSchema = Schema.from_dict({
            'data': fields.Raw(),
        })
        OutputSchema = datatypes.get('Y') or datatypes.get('result') or DefaultOutputSchema
        self.render_operations(InputSchema, {200: OutputSchema})
        self.add_schema_from_datatype(InputSchema.__name__, InputSchema)
        self.add_schema_from_datatype(OutputSchema.__name__, OutputSchema)

    def render_operations(self, input_schema, responses):  # add operations
        # renders paths according to self.path_template and signature.actions
        # if no signature.actions are specified, uses self.default_actions
        # inputs: Schema
        # responses: mapping http status => schema, or Schema => 200: Schema
        name = self.name
        resource = self.resource
        responses = {200: responses} if isinstance(responses, Schema) else dict(responses)
        operations = []
        actions = self.signature.get('actions') or self.default_actions
        for action, http_methods in actions.items():
            for http_method in http_methods:
                operation_id = f'{self.name}#{action}'
                opspec = (Operation(self.spec, http_method, operation_id=operation_id)
                          .add_request(input_schema))
                for status_code, output_schema in responses.items():
                    opspec.add_response(status_code, output_schema)
                operations.append(opspec)
            self.spec.path(self.path_template.format(**locals()),
                           operations=Operation.to_path(operations))

    def datatypes(self, orient):
        # convert signature specs back to Schema types
        signature = self.signature
        store = self.store
        name = self.name.replace('/', '_')
        if signature:
            datatypes = {
                k: store._datatype_from_schema(spec['schema'], name=f'{name}_{k}', orient=orient)
                for k, spec in signature.items() if k in ('X', 'Y', 'result') and spec
            }
        else:
            datatypes = None
        if not datatypes:
            raise ValueError(f"Missing signature specification for object {store.prefix}/{name}")
        return datatypes


class SpecFromModelHelper(SpecFromResourceHelperBase):
    default_actions = {'predict': ['post']}
    default_orient = 'records'

    def datatypes(self, orient):
        name = self.name.replace('/', '_')
        datatypes_ = super().datatypes(orient)
        datatypes_['X'] = Schema.from_dict({
            'data': self.match_data_orient(datatypes_.get('X', fields.Raw), orient),
            'columns': fields.List(fields.String),
            'shape': fields.List(fields.Integer),
        }, name=f'PredictInput_{name}')
        Result = datatypes_.get('Y') or datatypes_.get('result')
        datatypes_['Y'] = Schema.from_dict({
            'model': fields.String(),
            'result': fields.Nested(Result) if Result else fields.Raw(),
            'resource_uri': fields.String(),
        }, name=f'PredictOutput_{name}')
        return datatypes_


class SpecFromDatasetHelper(SpecFromResourceHelperBase):
    path_template = '/api/v1/dataset/{name}'
    default_actions = {'get': ['get'], 'put': ['put']}

    def datatypes(self, orient):
        name = self.name.replace('/', '_')
        store = self.store
        meta = self.meta
        datasetType = store._datatype_from_metadata(meta.to_dict(), orient=orient)
        DatasetInput = Schema.from_dict({
            'data': self.match_data_orient(datasetType, orient),
            'dtypes': fields.Raw(),
            'append': fields.Boolean(),
        }, name=f'DatasetInput_{name}')
        datatypes_ = {
            'X': DatasetInput,
            'result': DatasetInput,
        }
        return datatypes_


class SpecFromScriptHelper(SpecFromResourceHelperBase):
    path_template = '/api/v1/{resource}/{name}/{action}'
    default_actions = {'run': ['post']}
    default_orient = 'records'

    def datatypes(self, orient):
        datatypes_ = super().datatypes(orient)
        name = self.name.replace('/', '_')
        Result = datatypes_.get('result')
        datatypes_['result'] = Schema.from_dict({
            'resource_uri': fields.String(),
            'script': fields.String(),
            'result': Result if Result else fields.Raw,
            'runtimes': fields.Float,
            'started': fields.DateTime,
            'ended': fields.DateTime,
        }, name=f'ScriptOutput_{name}')
        return datatypes_


class SpecFromServiceHelper(SpecFromResourceHelperBase):
    path_template = '/api/service/{name}'
    default_actions = {'run': ['post'], 'predict': ['post']}
    default_orient = 'records'

    def datatypes(self, orient):
        try:
            datatypes = super().datatypes(orient=orient)
        except ValueError:
            if self.store.prefix == 'data/':
                warnings.warn('datasets are not properly supported as services yet, '
                              'please use the generic /api/v1/dataset resource')
                datasetHelper = SpecFromDatasetHelper(self.meta, self.name, self.resource,
                                                      self.signature, self.spec, self.store)
                datatypes = datasetHelper.datatypes(orient=orient)
            else:
                raise
        return datatypes




import re

from omegaml.backends.basedata import BaseDataBackend
from omegaml.backends.mlflow.localprojects import MLFlowProject
from omegaml.backends.package.packager import RunnablePackageMixin


class MLFlowGitProjectBackend(RunnablePackageMixin, BaseDataBackend):
    """
    Backend to support git-sourced MLFlow projects

    This supports any git remote source that MLFlow supports

    Usage::

        om.scripts.put('mlflow+ssh://git@github.com/<user>/<repo>.git', '<pkgname>')
        om.scripts.put('mlflow+https://git@github.com/<user>/<repo>.git', '<pkgname>')
        om.scripts.put('https://github.com/<user>/<repo>', '<pkgname>', kind='mlflow.project')

    See Also:
        * https://packaging.python.org/tutorials/packaging-projects/
        * https://docs.github.com/en/get-started/getting-started-with-git/managing-remote-repositories#changing-a-remote-repositorys-url
    """
    KIND = 'mlflow.gitproject'
    MLFLOW_GIT_PREFIX = re.compile(r'^mlflow\+(git|ssh|https).*')
    GIT_PREFIX = 'https'

    @classmethod
    def supports(self, obj, name, **kwargs):
        is_mlflow_git = isinstance(obj, str) and self.MLFLOW_GIT_PREFIX.match(obj)
        is_git = isinstance(obj, str) and obj.startswith(self.GIT_PREFIX) and kwargs.get('kind') == self.KIND
        return is_mlflow_git or is_git

    def put(self, obj, name, attributes=None, **kwargs):
        """
        save a MLFlow git-sourceable project
        """
        git_uri = obj.replace('mlflow+', '')
        return self.data_store._make_metadata(
            name=name,
            prefix=self.data_store.prefix,
            bucket=self.data_store.bucket,
            kind=MLFlowGitProjectBackend.KIND,
            uri=git_uri,
            attributes=attributes).save()

    def get(self, name, **kwargs):
        """
        Load MLFlow project from git uri

        :return: the loaded module
        """
        meta = self.data_store.metadata(name)
        mod = MLFlowProject(meta.uri)
        return mod



from omegaml.backends.restapi.job import GenericJobResource
from omegaml.backends.restapi.model import GenericModelResource
from omegaml.backends.restapi.script import GenericScriptResource


class GenericServiceResource:
    """ backend for /api/service/ resources

    Acts as an adapter to a concrete resource, returning only its bare result.
    If the resource (=stored object) has a signature attached, it has already
    been validated by the object's backend or as part of the SignatureMixin.

    If your resource returns a value that is not json serializable (i.e. the
    return value is not a dict), dict(value=<result>) will be returned.
    """
    RESOURCE_MAP = {
        'models/': GenericModelResource,
        'scripts/': GenericScriptResource,
        'jobs/': GenericJobResource,
    }

    def __init__(self, om, is_async=False):
        self.om = om
        self.is_async = is_async

    def doc(self, resource_id, query, payload):
        om = self.om
        meta = self.resolve_object(resource_id)
        specs = om.runtime.swagger(include=meta.name, format='dict', as_service=True)
        return specs

    def predict(self, model_id, query, payload):
        # TODO the validation should be in SignatureMixin (as with scripts), however
        #      this would not match the semantics of runtime ModelProxy/ModelBackend, whereby ModelProxy
        #      stores the payload as a dataset and only passes the dataset name for the ModelBackend to retrieve
        #      Options
        #      1) keep like this - mismatch, validation only works through the service call
        #      2) move all validation to service call - at odds with pre/post call semantics of Backend.perform()
        #      3) refactor model data retrieval to SignatureMixin, away from model backend (or optional)
        #      4) move validation responsibility to backend - not good, removes generality of validation
        #      5) change ModelProxy semantics, i.e. do not store payload as a dataset, but pass as actual payload
        #         (requires a ModelBackend/Mixin.pre_predict to keep semantics with backend)
        self.om.models.validate(model_id, X=payload)
        promise = self.om.runtime.model(model_id).predict(payload, **query)
        result = self.prepare_result(promise.get(), model_id=model_id) if not self.is_async else promise
        self.om.models.validate(model_id, Y=result) if isinstance(result, dict) else None
        return result

    def predict_proba(self, model_id, query, payload):
        self.om.models.validate(model_id, X=payload)
        promise = self.om.runtime.model(model_id).predict(payload, **query)
        result = self.prepare_result(promise.get(), model_id=model_id) if not self.is_async else promise
        self.om.models.validate(model_id, Y=result) if isinstance(result, dict) else None
        return result

    def default_action(http_method):
        # decorator to generate a method for each http method
        def do_action(self, resource_id, query, payload):
            meta = self.resolve_object(resource_id)
            # check first method that matches http_method in signature.actions
            actions = meta.attributes.get('signature', {}).get('actions') or []
            for method, http_methods in actions:
                # SignatureMixin.link_datatype
                if http_method in http_methods:
                    meth = getattr(self, method)
                    break
            else:
                # no matching signature.actions found, revert to default for object type
                if meta.prefix == 'models/':
                    meth = self.predict
                elif meta.prefix in ['scripts/', 'jobs/']:
                    meth = self.run
                else:
                    raise ValueError(f'No default action for {meta.prefix}/{resource_id}')
            result = meth(resource_id, query, payload)
            return result

        return do_action

    get = default_action('get')
    post = default_action('post')
    put = default_action('put')
    delete = default_action('delete')

    def run(self, script_id, query, payload):
        promise = self.om.runtime.script(script_id).run(payload, __format='python', **query)
        result = self.prepare_result(promise.get(), script_id=script_id) if not self.is_async else promise
        return result

    def prepare_result(self, result, model_id=None, script_id=None, resource_name=None, **kwargs):
        if script_id or resource_name:
            # FIXME all resource outputs provide 'result', except jobs
            # scripts provide structured task output, not just the result
            data = result.get('job_results') or result.get('result')
        else:
            data = result
        # ensure response is json serializable
        if not isinstance(data, dict):
            data = {'data': data}
        return data

    def resolve_object(self, resource_id):
        objects = self.om.list(resource_id, raw=True)
        meta = objects[-1] if objects else None
        if not meta or meta.prefix not in self.RESOURCE_MAP:
            raise ValueError(f'{resource_id} is not available as a service')
        return meta


class GenericServiceResourceFallback(GenericServiceResource):
    """ backend for /api/service/ resources

    Acts as an adapter to a concrete resource, returning only its bare result.
    If the resource (=stored object) has a signature attached, it has already
    been validated by the object's backend or as part of the SignatureMixin
    """
    RESOURCE_MAP = {
        'models/': GenericModelResource,
        'scripts/': GenericScriptResource,
        'jobs/': GenericJobResource,
    }

    def __getattr__(self, item):
        for resource in self.RESOURCE_MAP.values():
            if hasattr(resource, item):
                return self.handle_resource_method(item)
        raise AttributeError(item)

    def handle_resource_method(self, method):
        # returns a wrapper to forward the resource method to the actual resource
        def resource_method_adapter(resource_pk, query, payload):
            # find the actual object, get the actual Resource to handle it
            meta = self.resolve_object(resource_pk)
            Resource = self.RESOURCE_MAP[meta.prefix]
            # this behaves the same way as a call on the actual resource
            # -- except we intercept Resource.prepare_result_from to provide result extraction
            resource = Resource(self.om, is_async=self.is_async)
            self.handle_resource_prepare_result(resource)
            meth = getattr(resource, method)
            return meth(resource_pk, query, payload)

        return resource_method_adapter

    def handle_resource_prepare_result(self, resource):
        # intercepts resource.prepare_result to remove the default api outputs
        def override(actual_prepare_result):
            def prepare_result_adapter(*args, **kwargs):
                result = actual_prepare_result(*args, **kwargs)
                # FIXME all resource outputs provide 'result', except jobs
                data = result.get('job_results') or result.get('result')
                if not isinstance(data, dict):
                    # ensure response is json serializable
                    data = {'data': data}
                return data

            return prepare_result_adapter

        resource.prepare_result = override(resource.prepare_result)

    def prepare_result(self, result, resource_name=None, **kwargs):
        meta = self.resolve_object(resource_name)
        Resource = self.RESOURCE_MAP[meta.prefix]
        resource = Resource(self.om, is_async=self.is_async)
        self.handle_resource_prepare_result(resource)
        return resource.prepare_result(result, **kwargs)



""" omega-ml bulk deployment utility
(c) 2020 one2seven GmbH, Switzerland

Enables deployment of datasets, models, scripts, jobs as well as cloud
resources from a single configuration file. This is currently a separate
utility that will be integrated into the omega-ml cli.

Usage:
    To deploy a complete application, create the deploy.yml as per below
    and run:

    # initial deployment
    $ om runtime deploy --action add

    # subsequent deployments
    $ om runtime deploy

    # select specific parts to deploy
    $ python omdeploy.py --select scripts.apps/helloworld

    # dry run, show commands, do not exeucte
    $ python omdeploy.py --dry

    # deploy.yml (example showing most options, specify required parts only)
      # -- syntax is the equivalent of the om cli, where each command is
      #    a map of <arg>=<value> to specify the command parameters, and where
      #    the command is applied according to the omdeploy --action value
      # -- for example, the datasets entry corresponds to
      #    om datasets put data/mydata.csv mydata
      # -- additional commands include kubectl and shell, which both take
      #    a command: <cmd> entry
      # -- each command can also specify a dependency or a sequence which
      #    denotes the execution sequence
      #    ... depends: scripts => execute only when all scripts have been deployed
      #    ... sequence:
      # -- env variables can be specified as {VARIABLE} in any leaf value
      #    in-file variables can be centralized in a vars: section on top and
      #    referenced the same way, e.g. {var}. Convention: env vars in uppercase,
      #    in-file vars in lowercase
      vars:
         - foo=bar
      datasets:
         - name: mydata
           local: data/mydata.csv
      models:
         - name: mymodel
           local: package.mymodel
      scripts:
         - name: apps/helloworld
           local: ./helloworld
         - name: apps/myservice
           local: {om-baseapp} # om-baseapp is a placeholder for a dummy app
           metadata:
              appdef:
                 image: myimage
                 command: '"/bin/bash", "-c", "start.sh"'
      runtime:
         - action: models
           kind: fit
           name: mymodel
           options: mydata[^y] mydata[y]
         - action: restart
           kind: app
           name: helloworld
         - action: restart
           kind: app
           name: myservice
      cloud:
         # cloud is available in commercial edition only
         - kind: appingress
           specs:
             appname: helloworld
             hostname: www.mydomain.com
           depends: scripts
      kubectl:
          - command: apply -f configmap.yml
      shell:
          - command: curl -v https://www.mydomain.com
            sequence: 9999

"""
import logging

import argparse
import omegaml as om
import os
import re
import subprocess
import yaml
from omegaml.client import cli

parser = argparse.ArgumentParser(description='omegaml scripted deploy')
parser.add_argument('--file', dest='deployfile', default='deploy.yml',
                    help='/path/to/deploy.yml')
parser.add_argument('--dry', default=False, action='store_true')
parser.add_argument('--action', default='update',
                    help='add, update, remove')
parser.add_argument('--select', default='',
                    help='subset of assets to apply')
parser.add_argument('--specs', default='',
                    help='comma separated list of <var>=<value>[,...]')

DIRECT_COMMANDS = ['kubectl', 'shell']
COMMAND_ORDER = 'cloud,shell,kubectl,datasets,scripts,runtime,appingress'
SEQUENCE_SPACING = 10
SPECS_CLI_MAP = {
    'scripts': 'scripts {action} {local} {name} {options}',
    'datasets': 'datasets {action} {local} {name} {options}',
    'models': 'models {action} {local} {name} {options}',
    'jobs': 'jobs {action} {local} {name} {options}',
    'runtime': 'runtime {kind} {action} {name} {options}',
    'cloud': 'cloud {action} {kind} --specs {specs} {options}',
    'kubectl': 'kubectl {command}',
    'shell': '{command}'
}
ACTION_MAP = {
    'update': {
        '_default_': 'put',
        'runtime': 'restart',
        'cloud': 'update',
    },
    'add': {
        '_default_': 'put',
        'runtime': 'restart',
        'cloud': 'add',
    },
    'remove': {
        '_default_': 'drop',
        'runtime': 'status',
        'cloud': 'remove',
    },
}

METADATA_TYPES = ('scripts', 'datasets', 'jobs', 'models')
DEFAULT_VARS = {
    'om-baseapp': "git+https://github.com/omegaml/apps.git#subdirectory=helloworld&egg=helloworld",
}

module_logger = logging.getLogger(__name__)


def process(specs_file, action='plan', dry=False, select=None, specs=None, cli_logger=None):
    logger = cli_logger or module_logger
    order = COMMAND_ORDER
    commands = []
    vars = dict(action=action, dry=dry, select=select)
    vars.update({**os.environ, **DEFAULT_VARS})
    selected = (select or '').split(',')
    specs = dict([pair.strip().split('=', 1) for pair in specs.split(',')]) if specs else {}

    def render_vars(d, _doublepass=True, **vars):
        for k, v in d.items():
            if isinstance(v, dict):
                render_vars(v, **vars)
            elif isinstance(v, str):
                # remove new lines because cli commands never contain new lines
                d[k] = v.format(**vars).replace('\n', '')
        if _doublepass:
            render_vars(d, **vars, _doublepass=False)

    def prepare(cmd, item):
        if 'specs' in item:
            item['specs'] = ','.join(f'{k}={v}'
                                     for k, v in item['specs'].items())
        if action in ACTION_MAP:
            default_action = ACTION_MAP[action].get(cmd, ACTION_MAP[action].get('_default_'))
        else:
            default_action = action
        item.setdefault('action', default_action)
        item.setdefault('options', '')
        item.setdefault('local', '')
        item.setdefault('kind', '')
        clicmd = SPECS_CLI_MAP[cmd]
        command = {
            'cmd': cmd,
            'clicmd': clicmd,
            'item': item,
            'depends': item.get('depends'),
            'sequence': item.get('sequence', (len(commands) + 1) * SEQUENCE_SPACING),
            'metadata': item.get('metadata'),
        }
        try:
            action_override = {'action': item['action']} if not str(item.get('action')).startswith('{') else {}
            render_vars(command, **{**item, **vars, **specs, **action_override})
        except KeyError as e:
            logger.error(f"Variable {e} must be set in {cmd} {item}")
            exit(1)
        commands.append(command)

    def apply_meta(cmd):
        meta = cmd.get('metadata')
        if meta and cmd['cmd'] in METADATA_TYPES:
            if dry:
                logger.info(f'DRY: metadata update {meta}')
            else:
                store = getattr(om, cmd['cmd'])
                s_meta = store.metadata(cmd['item']['name'])
                s_meta.attributes.update(meta)
                s_meta.save()

    def is_selected(cmd):
        # find items to process
        # -- if selected is specified
        # -- get values for id, name, or kind in each cmd item
        # -- match against selected items (each is a string or re)
        # -- example:
        #    selected=['kubectl.def']
        #    select_keys = ['def', 'xyz'] # id: def, name: xyz
        #    lookup = 'kuebctl.abc', 'kubectl.def', 'kubectl.xyz']
        #    re.match('kubectl.def', 'kubectl.def') => True
        select_from_keys = ['id', 'name', 'kind']
        select_keys = [v for v in [cmd['item'].get(k) for k in select_from_keys] if v]
        select_key = select_keys[0] if select_keys else '*'
        lookup = '.'.join([v for v in (cmd.get('cmd'), select_key) if v])
        return lookup, selected and any(re.match(s, lookup) for s in selected)

    def apply():
        sequenced = sorted(commands,
                           key=lambda v: (len(commands) + 1) * SEQUENCE_SPACING if v.get('depends') else v.get(
                               'sequence'))
        for cmd in sequenced:
            lookup, should_process = is_selected(cmd)
            if not should_process:
                logger.info(f"ignoring {lookup} because not in {selected}")
                continue
            if cmd['cmd'] not in DIRECT_COMMANDS:
                # om cli
                if dry:
                    logger.info(f"DRY: om {cmd['clicmd']}")
                else:
                    logger.info(f"om {cmd['clicmd']}")
                    argv = [v for v in cmd['clicmd'].split(' ') if v]
                    cli.main(argv=argv, logger=logger)
                apply_meta(cmd)
            else:
                # shell
                shellcmd = cmd['clicmd']
                if dry:
                    logger.info(f"DRY: {shellcmd} ")
                else:
                    logger.info(f'{shellcmd}')
                    result = subprocess.run(shellcmd, shell=True, capture_output=True)
                    logger.info(result.stdout)

    def load():
        with open(specs_file) as fin:
            deploy_specs = yaml.safe_load(fin)
            try:
                vars_update = {k: v.format(**vars)
                               for k, v in deploy_specs.get('vars', {}).items()}
            except KeyError as e:
                logger.info(f"Variable {e} must be set in vars section")
                exit(1)
            vars.update(vars_update)
            for cmd in order.split(','):
                for item in deploy_specs.get(cmd, []):
                    prepare(cmd, item)

    load()
    apply()


if __name__ == '__main__':
    args = parser.parse_args()
    process(args.deployfile, dry=args.dry, action=args.action, select=args.select, specs=args.specs)



import os

import mlflow

from omegaml.backends.basemodel import BaseModelBackend


class MLFlowRegistryBackend(BaseModelBackend):
    """
    Backend to support registry-sourced MLFlow projects

    This supports any model tracking URI format that MLFlow supports

    Usage:
        om.scripts.put('mlflow+models://<model-name>/<runid>', '<model-name>')


    See Also
        https://www.mlflow.org/docs/latest/tracking.html
    """
    KIND = 'mlflow.registrymodel'
    MLFLOW_MODELS_PREFIX = 'mlflow+models'

    @classmethod
    def supports(self, obj, name, **kwargs):
        is_mlflow_models = isinstance(obj, str) and obj.startswith(self.MLFLOW_MODELS_PREFIX)
        return is_mlflow_models

    def put(self, obj, name, tracking_uri=None, attributes=None, **kwargs):
        """ save a MLFlow tracking-sourceable project by storing the models uri in meta.uri
        """
        models_uri = obj.replace('mlflow+models://', 'models:/')
        meta = self.data_store._make_metadata(
            name=name,
            prefix=self.data_store.prefix,
            bucket=self.data_store.bucket,
            kind=MLFlowRegistryBackend.KIND,
            uri=models_uri,
            attributes=attributes)
        self._ensure_tracking_uri(meta, tracking_uri=tracking_uri)
        return meta.save()

    def get(self, name, tracking_uri=None, **kwargs):
        """ Load MLFlow model from the tracking uri stored in metadata.uri """
        meta = self.data_store.metadata(name)
        tracking_uri = tracking_uri or meta.kind_meta['tracking_uri']
        self._ensure_tracking_uri(meta, tracking_uri=tracking_uri)
        model = mlflow.pyfunc.load_model(model_uri=f"{meta.uri}")
        return model

    def _ensure_tracking_uri(self, meta, tracking_uri=None):
        tracking_uri = (tracking_uri or mlflow.get_tracking_uri() or meta.kind_meta.get('tracking_uri') or
                        os.environ.get('MLFLOW_TRACKING_URI'))
        meta.kind_meta['tracking_uri'] = tracking_uri
        assert tracking_uri, "pass tracking_uri= or set kind_meta.tracking_uri or set env MLFLOW_TRACKING_URI"
        mlflow.tracking.MlflowClient(tracking_uri).list_registered_models(max_results=1)
        mlflow.set_tracking_uri(tracking_uri)




import getpass

import sys
from celery import Task
from contextlib import contextmanager
from kombu.serialization import registry
from kombu.utils import cached_property

from omegaml.client.auth import AuthenticationEnv
from omegaml.store.logging import OmegaLoggingHandler


class EagerSerializationTaskMixin(object):
    # ensure eager tasks are being serialized to capture serialization errors
    # adopted from https://github.com/celery/celery/issues/4008#issuecomment-330292405
    abstract = True

    def _not_eager(self):
        app = self._app
        return not app.conf.CELERY_ALWAYS_EAGER

    @cached_property
    def _task_serializer(self):
        app = self._app
        return app.conf.CELERY_TASK_SERIALIZER

    def apply_async(self, args=None, kwargs=None, *args_, **kwargs_):
        if self._not_eager:
            return super(EagerSerializationTaskMixin, self).apply_async(args=args, kwargs=kwargs, *args_, **kwargs_)
        # only execute if eager
        # -- perform a serialization / deserialization roundtrip, as we would in distributed mode (not eager)
        sargs, skwargs = self._eager_serialize_args(args=args, kwargs=kwargs, **kwargs_)
        sargs_, skwargs_ = self._eager_serialize_args(args=args_, kwargs=kwargs_, **kwargs_)
        # -- call actual task with deserialized args, kwargs, as it would be by remote
        result = super(EagerSerializationTaskMixin, self).apply_async(args=sargs, kwargs=skwargs, *sargs_, **skwargs_)
        # -- do the same for the result
        result = self._eager_serialize_result(result, **kwargs_)
        return result

    def _eager_serialize_args(self, args=None, kwargs=None, **kwargs_):
        # Perform a noop serialization backtrip to assert args and kwargs
        # will be serialized appropriately when an async call through kombu
        # is actually performed. This is done to make sure we catch the
        # serializations errors with our test suite which runs with the
        # CELERY_ALWAYS_EAGER setting set to True. See the following Celery
        # issue for details https://github.com/celery/celery/issues/4008.
        app = self._app
        producer = kwargs_.get('producer') if kwargs else None
        with app.producer_or_acquire(producer) as producer:
            serializer = kwargs_.get('serializer', producer.serializer) or self._task_serializer
            registry.enable(serializer)
            args_content_type, args_content_encoding, args_data = registry.dumps(args, serializer)
            kwargs_content_type, kwargs_content_encoding, kwargs_data = registry.dumps(kwargs, serializer)
            args = registry.loads(args_data, args_content_type, args_content_encoding)
            kwargs = registry.loads(kwargs_data, kwargs_content_type, kwargs_content_encoding)
        return args, kwargs

    def _eager_serialize_result(self, result, **kwargs_):
        app = self._app
        producer = kwargs_.get('producer') if kwargs_ else None
        result_value = result._result
        with app.producer_or_acquire(producer) as producer:
            serializer = kwargs_.get('serializer', producer.serializer) or self._task_serializer
            registry.enable(serializer)
            dtype, encoding, data = registry.dumps(result_value, serializer)
            result._result = registry.loads(data, dtype, encoding)
        return result


class OmegamlTask(EagerSerializationTaskMixin, Task):
    abstract = True

    def __init__(self, *args, **kwargs):
        super(OmegamlTask, self).__init__(*args, **kwargs)
        self.auth_env = AuthenticationEnv.active()

    @property
    def om(self):
        # TODO do some more intelligent caching, i.e. by client/auth
        if not hasattr(self.request, '_om'):
            self.request._om = None
        if self.request._om is None:
            bucket = self.system_kwargs.get('__bucket')
            auth = self.system_kwargs.get('__auth')
            self.request._om = self.auth_env.get_omega_for_task(self, auth=auth)[bucket]
        return self.request._om

    def get_delegate(self, name, kind='models', pass_as='model_store'):
        get_delegate_provider = getattr(self.om, kind)
        self.enable_delegate_tracking(name, kind, get_delegate_provider)
        kwargs = dict(data_store=self.om.datasets,
                      tracking=self.tracking)
        kwargs[pass_as] = get_delegate_provider
        result = get_delegate_provider.get_backend(name, **kwargs)
        return result

    def enable_delegate_tracking(self, name, kind, delegate_provider):
        exp = self.tracking.experiment
        meta = delegate_provider.metadata(name)
        exp.log_artifact(meta, 'related')
        tracking = meta.attributes.setdefault('tracking', {})
        tracking['experiments'] = set(tracking.get('experiments', []) + [exp._experiment])
        meta.save()
        return meta

    @property
    def current_userid(self):
        return getattr(self.om.runtime.auth, 'userid', getpass.getuser())

    @property
    def delegate_args(self):
        return self.request.args

    @property
    def delegate_kwargs(self):
        kwargs = self.request.kwargs or {}
        return {k: v for k, v in kwargs.items() if not k.startswith('__')}

    @property
    def system_kwargs(self):
        kwargs = self.request.kwargs or {}
        return {k: v for k, v in kwargs.items() if k.startswith('__')}

    @property
    def tracking(self):
        if not hasattr(self.request, '_om_tracking'):
            self.request._om_tracking = None
        if self.request._om_tracking is None:
            kwargs = self.request.kwargs or {}
            experiment = kwargs.get('__experiment')
            if experiment is not None:
                # we reuse implied_run=False to use the currently active run,
                # i.e. with block will NOT call exp.start() if an experiment is already running
                tracker = self.om.runtime.experiment(experiment, implied_run=False)
                tracker.log_extra(taskid=self.request.id, userid=self.current_userid)
                self.request._om_tracking = tracker
            else:
                self.request._om_tracking = self.om.runtime.experiment('.notrack', provider='notrack')
        return self.request._om_tracking

    @property
    def logging(self):
        kwargs = self.request.kwargs or {}
        logging = kwargs.get('__logging', False)
        if isinstance(logging, tuple):
            logname, level = logging
            if logname is True:
                logname = 'root'
        elif isinstance(logging, str):
            if logging.lower() in ('info', 'warn', 'fatal', 'error', 'debug', 'critical'):
                logname, level = 'root', logging.upper()
            else:
                logname, level = logging, 'INFO'
        elif logging is True:
            logname, level = 'root', 'INFO'
        else:
            logname, level = None, 'NOTSET'
        return logname, level

    def __call__(self, *args, **kwargs):
        import logging

        @contextmanager
        def task_logging():
            logname, level = self.logging
            if logname:
                self.om.logger.setLevel(level)
                logger = logging.getLogger(name=logname if logname != 'root' else None)
                logger.setLevel(level)
                save_stdout, save_stderr = sys.stdout, sys.stderr
                self.app.log.redirect_stdouts_to_logger(logger, loglevel=level)
                handler = OmegaLoggingHandler.setup(store=self.om.datasets,
                                                    logger=logger,
                                                    exit_hook=True,
                                                    level=level)
            else:
                logger = None
            try:
                yield
            finally:
                if logger:
                    handler.flush()
                    handler.close()
                    logger.removeHandler(handler)
                    # reset stdout redirects
                    sys.stdout, sys.stderr = save_stdout, save_stderr

        with task_logging():
            # start experiment block, if required
            # -- this ensures we call exp.start() / exp.stop() if the task was started in context
            #    of this celery task
            with self.tracking as exp:
                # log task call event
                # -- only log delegate args, kwargs
                # -- avoid logging internal arguments
                exp.log_event(f'task_call', self.name, {'args': self.delegate_args,
                                                        'kwargs': self.delegate_kwargs})
                if self.request.id is not None:
                    # PERFTUNED
                    # if we have a request, avoid super().__call__()
                    # calling super() removes the request.id, which means our request._om caching does not work
                    # Rationale: https://github.com/celery/celery/issues/2633#issuecomment-286694356
                    result = self.run(*args, **self.delegate_kwargs)
                else:
                    # PERFISSUE new task cannot access the request cache, new omega instance is created
                    # if we don't have a request.id it means we don't have a proper callstack
                    # so we need to create one using super().__call__
                    # this may happen when calling tasks directly, such as in run_omega_callback_script
                    result = super().__call__(*args, **kwargs)
        return result

    def reset(self):
        # ensure next call will start over and get a new om instance
        self.request._om = None

    def send_event(self, type, **fields):
        # ensure result masking in events
        if 'result' in fields:
            fields['result'] = AuthenticationEnv().active().resultauth(fields['result'])
        super().send_event(type, **fields)

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        with self.tracking as exp:
            exp.log_event(f'task_failure', self.name, {
                'exception': repr(exc),
                'task_id': task_id,
            })
            exp.log_extra(taskid=None, remove=True)
        self.reset()
        return super().on_failure(exc, task_id, args, kwargs, einfo)

    def on_retry(self, exc, task_id, args, kwargs, einfo):
        with self.tracking as exp:
            exp.log_event(f'task_retry', self.name, {
                'exception': repr(exc),
                'task_id': task_id,
            })
            exp.log_extra(taskid=None, remove=True)
        self.reset()
        return super().on_retry(exc, task_id, args, kwargs)

    def on_success(self, retval, task_id, args, kwargs):
        # on task success the experiment is stopped already (if we started it)
        exp = self.tracking
        exp.log_event(f'task_success', self.name, {
            'result': sanitized(retval),
            'task_id': task_id,
        })
        exp.log_extra(taskid=None, remove=True)
        self.reset()
        return super().on_success(retval, task_id, args, kwargs)


def get_dataset_representations(items):
    """
    returns dict with x and y datasets
    """
    results = {}
    results['Xname'] = items.get('Xname')
    results['Yname'] = items.get('Yname')
    return results


def sanitized(value):
    # fix because local Metadata object cannot be pickled
    if getattr(type(value), '__name__', None) == 'Metadata':
        value = repr(value)
    return value



from __future__ import absolute_import

import datetime
import glob
import os
import tempfile
import types
from shutil import rmtree
from zipfile import ZipFile, ZIP_DEFLATED

import joblib
from sklearn.base import BaseEstimator
from sklearn.model_selection import GridSearchCV

from omegaml.backends.basemodel import BaseModelBackend
from omegaml.documents import MDREGISTRY
from omegaml.util import reshaped, gsreshaped

# byte string
_u8 = lambda t: t.encode('UTF-8', 'replace') if isinstance(t, str) else t


class ScikitLearnBackendV1(BaseModelBackend):
    KIND = MDREGISTRY.SKLEARN_JOBLIB

    @classmethod
    def supports(self, obj, name, **kwargs):
        return isinstance(obj, BaseEstimator)

    # kept to support legacy scikit learn model serializations prior to ~scikit learn v0.18
    def _v1_package_model(self, model, filename):
        """
        Dumps a model using joblib and packages all of joblib files into a zip
        file
        """
        import joblib
        lpath = tempfile.mkdtemp()
        fname = os.path.basename(filename)
        mklfname = os.path.join(lpath, fname)
        zipfname = os.path.join(self.model_store.tmppath, fname)
        joblib.dump(model, mklfname, protocol=4)
        with ZipFile(zipfname, 'w', compression=ZIP_DEFLATED) as zipf:
            for part in glob.glob(os.path.join(lpath, '*')):
                zipf.write(part, os.path.basename(part))
        rmtree(lpath)
        return zipfname

    def _v1_extract_model(self, packagefname):
        """
        Loads a model using joblib from a zip file created with _package_model
        """
        import joblib
        lpath = tempfile.mkdtemp()
        fname = os.path.basename(packagefname)
        mklfname = os.path.join(lpath, fname)
        with ZipFile(packagefname) as zipf:
            zipf.extractall(lpath)
        model = joblib.load(mklfname)
        rmtree(lpath)
        return model

    def _v1_get_model(self, name, version=-1):
        """
        Retrieves a pre-stored model
        """
        filename = self.model_store._get_obj_store_key(name, '.omm')
        packagefname = os.path.join(self.model_store.tmppath, name)
        dirname = os.path.dirname(packagefname)
        try:
            os.makedirs(dirname)
        except OSError:
            # OSError is raised if path exists already
            pass
        meta = self.model_store.metadata(name, version=version)
        outf = self.model_store.fs.get_version(filename, version=version)
        with open(packagefname, 'wb') as zipf:
            zipf.write(meta.gridfile.read())
        model = self._v1_extract_model(packagefname)
        return model

    def _v1_put_model(self, obj, name, attributes=None):
        """
        Packages a model using joblib and stores in GridFS
        """
        zipfname = self._v1_package_model(obj, name)
        with open(zipfname, 'rb') as fzip:
            gridfile = self.model_store.fs.put(
                fzip, filename=self.model_store._get_obj_store_key(name, 'omm'))
        return self.model_store._make_metadata(
            name=name,
            prefix=self.model_store.prefix,
            bucket=self.model_store.bucket,
            kind=MDREGISTRY.SKLEARN_JOBLIB,
            attributes=attributes,
            gridfile=gridfile).save()


class ScikitLearnBackendV2(ScikitLearnBackendV1):
    """
    OmegaML backend to use with ScikitLearn
    """

    def _package_model(self, model, key, tmpfn):
        """
        Dumps a model using joblib and packages all of joblib files into a zip
        file
        """
        joblib.dump(model, tmpfn, protocol=4, compress=True)
        return tmpfn

    def _extract_model(self, infile, key, tmpfn):
        """
        Loads a model using joblib from a zip file created with _package_model
        """
        with open(tmpfn, 'wb') as pkgf:
            pkgf.write(infile.read())
        model = joblib.load(tmpfn)
        return model

    def get_model(self, name, version=-1):
        """
        Retrieves a pre-stored model
        """
        meta = self.model_store.metadata(name)
        if self._backend_version != meta.kind_meta.get(self._backend_version_tag):
            return super()._v1_get_model(name, version=version)
        return super().get_model(name, version=version)

    def put_model(self, obj, name, attributes=None, _kind_version=None):
        if _kind_version and _kind_version != self._backend_version:
            return super()._v1_put_model(obj, name, attributes=attributes)
        return super().put_model(obj, name, attributes=attributes)

    def predict(
          self, modelname, Xname, rName=None, pure_python=True, **kwargs):
        data = self.data_store.get(Xname)
        model = self.model_store.get(modelname)

        def store(result):
            if pure_python:
                result = result.tolist()
            if rName:
                meta = self.data_store.put(result, rName)
                result = meta
            return result

        result = process(maybe_chunked(model.predict,
                                       lambda data: as_args(reshaped(data)),
                                       data, **kwargs), fn=store, keep_last=True)
        return result

    def predict_proba(
          self, modelname, Xname, rName=None, pure_python=True, **kwargs):
        data = self.data_store.get(Xname)
        model = self.model_store.get(modelname)

        def store(result):
            if pure_python:
                result = result.tolist()
            if rName:
                meta = self.data_store.put(result, rName)
                result = meta
            return result

        result = process(maybe_chunked(model.predict_proba,
                                       lambda data: as_args(reshaped(data)),
                                       data, **kwargs), fn=store, keep_last=True)
        return result

    def fit(self, modelname, Xname, Yname=None, pure_python=True, **kwargs):
        model = self.model_store.get(modelname)
        X = self.data_store.get(Xname)
        Y = self.data_store.get(Yname) if Yname is not None else None
        model.fit(reshaped(X), reshaped(Y), **kwargs)
        meta = self.model_store.put(model, modelname)
        return meta

    def partial_fit(
          self, modelname, Xname, Yname=None, pure_python=True, **kwargs):
        model = self.model_store.get(modelname)
        X, metaX = self.data_store.get(Xname), self.data_store.metadata(Xname)
        Y, metaY = None, None
        if Yname:
            Y, metaY = (self.data_store.get(Yname),
                        self.data_store.metadata(Yname))
        process(maybe_chunked(model.partial_fit,
                              lambda X, Y: as_args(reshaped(X), reshaped(Y)),
                              X, Y, **kwargs))
        meta = self.model_store.put(model, modelname)
        return meta

    def score(
          self, modelname, Xname, Yname=None, rName=None, pure_python=True,
          **kwargs):
        model = self.model_store.get(modelname)
        X = self.data_store.get(Xname)
        Y = self.data_store.get(Yname)

        def store(result):
            if rName:
                meta = self.model_store.put(result, rName)
                result = meta
            return result

        result = process(maybe_chunked(model.score,
                                       lambda X, Y: as_args(reshaped(X), reshaped(Y)),
                                       X, Y, **kwargs), fn=store, keep_last=True)

        self.tracking.log_metric('score', result)
        return result

    def fit_transform(
          self, modelname, Xname, Yname=None, rName=None, pure_python=True,
          **kwargs):
        model = self.model_store.get(modelname)
        X, metaX = self.data_store.get(Xname), self.data_store.metadata(Xname)
        Y, metaY = None, None
        if Yname:
            Y, metaY = (self.data_store.get(Yname),
                        self.data_store.metadata(Yname))

        def store(result):
            if pure_python:
                result = result.tolist()
            if rName:
                meta = self.data_store.put(result, rName)
                result = meta
            return result

        result = process(maybe_chunked(model.fit_transform,
                                       lambda X, Y: as_args(reshaped(X), reshaped(Y)),
                                       X, Y, **kwargs), fn=store, keep_last=True)

        model_meta = self.model_store.put(model, modelname)
        return result if rName else model_meta

    def transform(self, modelname, Xname, rName=None, pure_python=True, **kwargs):
        model = self.model_store.get(modelname)
        X = self.data_store.get(Xname)

        def store(result):
            if pure_python:
                result = result.tolist()
            if rName:
                meta = self.data_store.put(result, rName)
                result = meta
            return result

        result = process(maybe_chunked(model.transform,
                                       lambda X: as_args(reshaped(X)),
                                       X, **kwargs), fn=store, keep_last=True)
        return result

    def decision_function(self, modelname, Xname, rName=None, pure_python=True, **kwargs):
        model = self.model_store.get(modelname)
        X = self.data_store.get(Xname)

        def store(result):
            if pure_python:
                result = result.tolist()
            if rName:
                meta = self.data_store.put(result, rName)
                result = meta
            return result

        result = process(maybe_chunked(model.decision_function,
                                       lambda X: as_args(reshaped(X)),
                                       X, **kwargs), fn=store, keep_last=True)
        return result

    def gridsearch(self, modelname, Xname, Yname=None, rName=None,
                   parameters=None, pure_python=True, **kwargs):
        model, meta = self.model_store.get(modelname), self.model_store.metadata(modelname)
        X = self.data_store.get(Xname)
        if Yname:
            y = self.data_store.get(Yname)
        else:
            y = None
        gs_model = GridSearchCV(cv=5, estimator=model, param_grid=parameters, **kwargs)
        gs_model.fit(X, gsreshaped(y))
        nowdt = datetime.datetime.now()
        if rName:
            gs_modelname = rName
        else:
            gs_modelname = '{}.{}.gs'.format(modelname, nowdt.isoformat())
        gs_meta = self.model_store.put(gs_model, gs_modelname)
        attributes = meta.attributes
        if not 'gridsearch' in attributes:
            attributes['gridsearch'] = []
        attributes['gridsearch'].append({
            'datetime': nowdt,
            'Xname': Xname,
            'Yname': Yname,
            'gsModel': gs_modelname,
        })
        meta.save()
        return meta


class ScikitLearnBackend(ScikitLearnBackendV2):
    pass


def process(it, fn=None, keep=False, keep_last=False):
    """
    consume iterator, optionally keeping results and calling a function

    Args:
        it: iterator to consume
        fn: optional, a function to call on result as result = fn(result)
        keep: optional, if True returns all results, False by default
        keep_last: optional, if True returns last result, False by default

    Returns:
        list of results if keep is True
        last result if keep_last is True
        None if keep and keep_last are False
    """
    results = [] if keep else None
    for result in it:
        result = fn(result) if fn else result
        results.append(result) if keep else None
    return result if keep_last else None


def maybe_chunked(meth, argfn, *args, **kwargs):
    """call a function with given args, kwargs, perhaps iterating over args

    Calls meth(*args, **kwargs) by the following rules:

    * if all values in *args are generators call meth() until generators exhausted
    * else run meth() once using *args as a single chunk
    * before each meth() call, argfn(*chunks) is called

    If all values in *args are generators this is equivalent to

        for values in zip(*args):
            yield meth(*values, **kwargs)

    If values in *args are not generators this is equivalent to

        yield meth(*args, **kwargs)

    Usage:
        X = get(...)
        Y = get(...)

        # original call
        model.partial_fit(X, Y, **kwargs)

        # transformed call, assuming X, Y can be generators over partial data
        # this will work whether X, Y are generators or not.
        for result in maybe_chunked(model.partial_fit,
                                    lambda *args: as_args(*args),
                                    X, Y, **kwargs):
            ... # process results as needed

        Note the (lambda ...), which is called on the X, Y args for each call
        on model.partial_fit. Use this to apply functions to each chunk, e.g.:

            # here the lambda receives the values of X, Y in each iteration
            maybe_chunked(...,
                          lambda X, Y: as_args(reshaped(X), reshaped(Y)),
                          ...)

        # combine with process() to simplify processing
        # essentiall this implements a for loop around maybe_chunked()
        resultl = process(maybe_chuncked(...), keep_last=True)

    Rationale:

        Avoid repeated code of the form

            X, Y = generators or actual values
            if isgenerator(X) and isgenerator(Y):
                for Xc, Yc in zip(X, Y):
                    chunk_result = meth(reshape(Xc), reshape(Yc), **kwargs)
                    # add logic to process chunk_result if needed
                    ...
                result = combine chunk_results, or keep only last
            else:
                result = meth(reshape(X), reshape(Y), **kwargs)

        Instead this can be written as

            result = process(maybe_chunked(meth,
                                           lambda X, Y: as_args(reshape(X), reshape(Y)),
                                           X, Y, **kwargs),
                             keep_last=True)

    Args:
        meth: function to call
        argfn: function to call on values for each iteration
        *args: arguments to meth
        **kwargs: kwargs to meth

    Returns:
        generator of meth() results
    """
    isgenerator = lambda v: isinstance(v, types.GeneratorType)
    if all(isgenerator(v) for v in args):
        for chunks in zip(*args):
            yield meth(*argfn(*chunks), **kwargs)
    else:
        yield meth(*argfn(*args), **kwargs)


def as_args(*args):
    # return the arguments as passed in
    # this is to enable e.g. lambda X: as_args(X), where X is an iterable
    # notice if we did not use as_args a subsequent fn(*(lambda X: X)) would
    # not be the same as fn(X), however fn(*(lambda X: as_args(X)) is equiv.
    return args



from os.path import basename, dirname

import base64
import json
import os
from shutil import make_archive, unpack_archive
from subprocess import run

from omegaml.backends.basedata import BaseDataBackend
from omegaml.backends.package.packager import RunnablePackageMixin
from omegaml.runtimes.rsystem import rhelper
from omegaml.util import tryOr


class RPackageData(RunnablePackageMixin, BaseDataBackend):
    """
    Backend to support R packages

    Usage::

        om.scripts.put('R://path/to/app.R', 'myname')
        om.scripts.get('myname')

    See Also:
        * Rscript
        * https://mastering-shiny.org/scaling-packaging.html (CC BY-NC-ND 4.0)
    """
    KIND = 'package.r'

    @classmethod
    def supports(self, obj, name, **kwargs):
        return isinstance(obj, str) and obj.startswith('R://')

    def put(self, obj, name, attributes=None, **kwargs):
        """
        save a R package

        This takes the full path to a setuptools setup.py, or a directory
        containing a setup.py file. It then executes `python setup.py sdist`
        and stores the resulting .tar.gz file in om.scripts

        :param obj: full path to package file or directory, syntax as
                    pkg://path/to/dist.tar.gz or pkg://path/to/setup.py
        :param name: name of package. must be the actual name of the package
                     as specified in setup.py
        :return: the Metadata object
        """
        pkgsrc = pkgdist = obj.split('//')[1]
        pkgsrc = pkgsrc.replace('app.R', '')
        if not 'tar.gz' in os.path.basename(pkgdist):
            distdir = os.path.join(pkgsrc, 'dist')
            os.makedirs(distdir, exist_ok=True)
            base_name = os.path.join(distdir, f'{name}')
            pkgdist = make_archive(base_name, 'gztar', pkgsrc)
        filename = self.data_store.object_store_key(name, 'pkg', hashed=True)
        gridfile = self._store_to_file(self.data_store, pkgdist, filename)
        return self.data_store._make_metadata(
            name=name,
            prefix=self.data_store.prefix,
            bucket=self.data_store.bucket,
            kind=RPackageData.KIND,
            attributes=attributes,
            gridfile=gridfile).save()

    def get(self, name, localpath=None, keep=False, install=True, **kwargs):
        """
        Load package from store, install it locally and load.

        This copies the package's .tar.gz file from om.scripts to a local temp
        path and runs `pip install` on it.

        :param name: the name of the package
        :param keep: keep the packages load path in sys.path, defaults to False
        :param localpath: the local path to store the package
        :param install: if True call pip install on the retrieved package
        :param kwargs:
        :return: the loaded module (RScript)
        """
        pkgname = basename(name)
        dstdir = localpath or self.data_store.tmppath
        packagefname = '{}.tar.gz'.format(os.path.join(localpath or self.packages_path, pkgname))
        os.makedirs(dirname(packagefname), exist_ok=True)
        meta = self.data_store.metadata(name)
        outf = meta.gridfile
        with open(packagefname, 'wb') as pkgf:
            pkgf.write(outf.read())
        if install:
            unpack_archive(packagefname, dstdir)
            mod = RScript(dstdir)
        else:
            mod = os.path.join(dstdir, pkgname)
        return mod

    @property
    def packages_path(self):
        return os.path.join(self.data_store.tmppath, 'packages')


class RScript:
    """ a Python proxy to the R process that runs a script

    This provides the ``mod.run()`` interface for scripts so that
    we can use the same semantics for R and python scripts.
    """
    def __init__(self, appdir):
        self.appdir = appdir

    def run(self, om, **kwargs):
        """ run the script in R session

        Usage:
            The script must exist as ``{self.appdir}/app.R``. It must implement
            the ``omega_run()`` function.

            Example::

                # app.R
                library(jsonlite)
                omega_run <- function(om, kwargs) {
                   # if om was passed as "0", this means we're in a local mode, i.e. must import omegaml
                   om <- if (om == "0") import("omegaml") else om
                   s <- fromJSON(rawToChar(base64_dec(kwargs)))
                   s$message <- "hello from R"
                   return(toJSON(s))
                }

        Notes:
            - If an R session is active, will run the script by calling the script's omega_run function
            - If no R session is active, will use RScript to source the script and run omega_run function
            - Expects the output to be in JSON format
        """
        r = rhelper()
        if r is None:
            r_kwargs = base64.b64encode(json.dumps(kwargs).encode('utf8')).decode('ascii')
            rcmd = fr'Rscript -e source("{self.appdir}/app.R") -e omega_run(0,"{r_kwargs}")'
            output = run(rcmd.split(' '), capture_output=True)
            output = output.stdout
        else:
            r.source(f'{self.appdir}/app.R')
            output = r.omega_run(om, kwargs)
        return tryOr(lambda: json.loads(output), output)



import json
import os
import unittest
from pathlib import Path

from omegaml import Omega
from omegaml.backends.rsystem.rscripts import RPackageData, RScript
from omegaml.runtimes.rsystem import r_available, inside_r
from omegaml.tests.util import OmegaTestMixin


@unittest.skipUnless(r_available or inside_r, "results consolidated as RSystemTestCase.test_inside_r result")
class RSystemScriptTests(OmegaTestMixin, unittest.TestCase):
    """ test running R functionality from within Python
    """

    def setUp(self):
        om = self.om = Omega()
        om.scripts.register_backend(RPackageData.KIND, RPackageData)
        self.clean()
        # force test mode for omega run from R
        os.environ['OMEGA_TEST_MODE'] = "yes"

    def _write_helloR(self):
        import omegaml
        om = self.om
        package_path = (Path(omegaml.__file__).parent / 'example' / 'demo' / 'helloR')
        meta = om.scripts.put(f'R://{package_path}/app.R', 'helloR')
        return meta

    def test_rpackage(self):
        """ test we can store and retrieve R packages """
        om = self.om
        meta = self._write_helloR()
        mod = om.scripts.get('helloR')
        self.assertIsInstance(mod, RScript)
        self.assertTrue(Path(mod.appdir).exists())
        self.assertTrue((Path(mod.appdir) / 'app.R').exists())
        output = mod.run(self.om, foo='bar')
        self.assertIsInstance(output, dict)
        self.assertIn('message', output)
        self.assertEqual(output['message'], ['hello from R'])
        self.assertEqual(output['scripts'], ['helloR'])

    def test_rpackage_runtime(self):
        """ test we can run R packages on the runtime """
        om = self.om
        meta = self._write_helloR()
        result = om.runtime.script('helloR').run().get()
        result = json.loads(result)
        self.assertIn('result', result)
        output = result['result']
        self.assertIn('message', output)
        self.assertEqual(output['message'], ['hello from R'])
        self.assertEqual(output['scripts'], ['helloR'])

